@@ -43,12 +43,6 @@ const char rc_copyright[] = "Copyright (c) 2007-2008 Roy Marples";
 #include <sys/utsname.h>
 #include <sys/wait.h>
 
-/* So we can coldplug net devices */
-#ifdef BSD
-# include <sys/socket.h>
-# include <ifaddrs.h>
-#endif
-
 #ifdef __linux__
 # include <asm/setup.h> /* for COMMAND_LINE_SIZE */
 #endif
@@ -79,7 +73,6 @@ const char rc_copyright[] = "Copyright (c) 2007-2008 Roy Marples";
 
 #define INITSH                  RC_LIBDIR "/sh/init.sh"
 #define INITEARLYSH             RC_LIBDIR "/sh/init-early.sh"
-#define HALTSH                  RC_INITDIR "/halt.sh"
 
 #define SHUTDOWN                "/sbin/shutdown"
 #define SULOGIN                 "/sbin/sulogin"
@@ -88,24 +81,22 @@ const char rc_copyright[] = "Copyright (c) 2007-2008 Roy Marples";
 
 #define DEVBOOT			"/dev/.rcboot"
 
-static char *RUNLEVEL = NULL;
-static char *PREVLEVEL = NULL;
-
 const char *applet = NULL;
-static char *runlevel = NULL;
-static RC_STRINGLIST *coldplugged_services = NULL;
-static RC_STRINGLIST *stop_services = NULL;
-static RC_STRINGLIST *start_services = NULL;
-static RC_STRINGLIST *types_n = NULL;
-static RC_STRINGLIST *types_nua = NULL;
-static RC_DEPTREE *deptree = NULL;
-static RC_HOOK hook_out = 0;
+static char *runlevel;
+static RC_STRINGLIST *hotplugged_services;
+static RC_STRINGLIST *stop_services;
+static RC_STRINGLIST *start_services;
+static RC_STRINGLIST *types_n;
+static RC_STRINGLIST *types_nua;
+static RC_DEPTREE *deptree;
+static RC_HOOK hook_out;
 
 struct termios *termios_orig = NULL;
 
 RC_PIDLIST service_pids;
 
-static void clean_failed(void)
+static void
+clean_failed(void)
 {
 	DIR *dp;
 	struct dirent *d;
@@ -120,13 +111,14 @@ static void clean_failed(void)
 			     (d->d_name[1] == '.' && d->d_name[2] == '\0')))
 				continue;
 
-			l = strlen(RC_SVCDIR "/failed/") + strlen(d->d_name) + 1;
+			l = strlen(RC_SVCDIR "/failed/") +
+				strlen(d->d_name) + 1;
 			path = xmalloc(sizeof(char) * l);
 			snprintf(path, l, RC_SVCDIR "/failed/%s", d->d_name);
 			if (path) {
 				if (unlink(path))
-					eerror("%s: unlink `%s': %s", applet, path,
-					       strerror(errno));
+					eerror("%s: unlink `%s': %s",
+					       applet, path, strerror(errno));
 				free(path);
 			}
 		}
@@ -134,50 +126,54 @@ static void clean_failed(void)
 	}
 }
 
-static void cleanup(void)
+static void
+cleanup(void)
 {
-	if (applet && strcmp(applet, "rc") == 0) {
-		RC_PID *p1 = LIST_FIRST(&service_pids);
-		RC_PID *p2;
+#ifdef DEBUG_MEMORY
+	RC_PID *p1 = LIST_FIRST(&service_pids);
+	RC_PID *p2;
+#endif
 
+	if (!rc_in_logger && !rc_in_plugin &&
+	    applet && strcmp(applet, "rc") == 0)
+	{
 		if (hook_out)
 			rc_plugin_run(hook_out, runlevel);
 
 		rc_plugin_unload();
 
-		if (! rc_in_plugin && termios_orig) {
+		if (termios_orig) {
 			tcsetattr(STDIN_FILENO, TCSANOW, termios_orig);
 			free(termios_orig);
 		}
 
-		while (p1) {
-			p2 = LIST_NEXT(p1, entries); 
-			free(p1);
-			p1 = p2;
-		}
-
-		rc_stringlist_free(coldplugged_services);
-		rc_stringlist_free(stop_services);
-		rc_stringlist_free(start_services);
-		rc_stringlist_free(types_n);
-		rc_stringlist_free(types_nua);
-		rc_deptree_free(deptree);
-
 		/* Clean runlevel start, stop markers */
-		if (! rc_in_plugin && ! rc_in_logger) {
-			rmdir(RC_STARTING);
-			rmdir(RC_STOPPING);
-			clean_failed();
-
-			rc_logger_close();
-		}
-
-		free(runlevel);
+		rmdir(RC_STARTING);
+		rmdir(RC_STOPPING);
+		clean_failed();
+		rc_logger_close();
 	}
+
+#ifdef DEBUG_MEMORY
+	while (p1) {
+		p2 = LIST_NEXT(p1, entries); 
+		free(p1);
+		p1 = p2;
+	}
+
+	rc_stringlist_free(hotplugged_services);
+	rc_stringlist_free(stop_services);
+	rc_stringlist_free(start_services);
+	rc_stringlist_free(types_n);
+	rc_stringlist_free(types_nua);
+	rc_deptree_free(deptree);
+	free(runlevel);
+#endif
 }
 
 #ifdef __linux__
-static char *proc_getent(const char *ent)
+static char *
+proc_getent(const char *ent)
 {
 	FILE *fp;
 	char proc[COMMAND_LINE_SIZE];
@@ -185,10 +181,10 @@ static char *proc_getent(const char *ent)
 	char *value = NULL;
 	int i;
 
-	if (! exists("/proc/cmdline"))
+	if (!exists("/proc/cmdline"))
 		return NULL;
 
-	if (! (fp = fopen("/proc/cmdline", "r"))) {
+	if (!(fp = fopen("/proc/cmdline", "r"))) {
 		eerror("failed to open `/proc/cmdline': %s", strerror(errno));
 		return NULL;
 	}
@@ -211,18 +207,19 @@ static char *proc_getent(const char *ent)
 }
 #endif
 
-static char read_key(bool block)
+static char
+read_key(bool block)
 {
 	struct termios termios;
 	char c = 0;
 	int fd = STDIN_FILENO;
 
-	if (! isatty(fd))
+	if (!isatty(fd))
 		return false;
 
 	/* Now save our terminal settings. We need to restore them at exit as we
 	 * will be changing it for non-blocking reads for Interactive */
-	if (! termios_orig) {
+	if (!termios_orig) {
 		termios_orig = xmalloc(sizeof(*termios_orig));
 		tcgetattr(fd, termios_orig);
 	}
@@ -236,15 +233,13 @@ static char read_key(bool block)
 		termios.c_cc[VTIME] = 0;
 	}
 	tcsetattr(fd, TCSANOW, &termios);
-
 	read(fd, &c, 1);
-
 	tcsetattr(fd, TCSANOW, termios_orig);
-
 	return c;
 }
 
-static bool want_interactive(void)
+static bool
+want_interactive(void)
 {
 	char c;
 	static bool gotinteractive;
@@ -252,32 +247,26 @@ static bool want_interactive(void)
 
 	if (rc_yesno(getenv("EINFO_QUIET")))
 		return false;
-
-	if (PREVLEVEL &&
-	    strcmp(PREVLEVEL, "N") != 0 &&
-	    strcmp(PREVLEVEL, "S") != 0 &&
-	    strcmp(PREVLEVEL, "1") != 0)
-		return false;
-
-	if (! gotinteractive) {
+	if (!gotinteractive) {
 		gotinteractive = true;
 		interactive = rc_conf_yesno("rc_interactive");
 	}
-	if (! interactive)
+	if (!interactive)
 		return false;
-
 	c = read_key(false);
 	return (c == 'I' || c == 'i') ? true : false;
 }
 
-static void mark_interactive(void)
+static void
+mark_interactive(void)
 {
 	FILE *fp = fopen(INTERACTIVE, "w");
 	if (fp)
 		fclose(fp);
 }
 
-static void run_program(const char *prog)
+static void
+run_program(const char *prog)
 {
 	struct sigaction sa;
 	sigset_t full;
@@ -322,31 +311,25 @@ static void run_program(const char *prog)
 		eerrorx("%s: failed to exec `%s'", applet, prog);
 }
 
-static void sulogin(bool cont)
+static void
+sulogin(bool cont)
 {
 #ifdef __linux__
 	const char *sys = rc_sys();
 
 	/* VSERVER and OPENVZ systems cannot do a sulogin */
-	if (sys && (strcmp(sys, "VSERVER") == 0 || strcmp(sys, "OPENVZ") == 0)) {
+	if (sys &&
+	    (strcmp(sys, "VSERVER") == 0 || strcmp(sys, "OPENVZ") == 0))
+	{
 		execl("/sbin/halt", "/sbin/halt", "-f", (char *) NULL);
 		eerrorx("%s: unable to exec `/sbin/halt': %s",
 			applet, strerror(errno));
 	}
 #endif
-
-	if (! cont) {
+	if (!cont) {
 		rc_logger_close();
-#ifdef __linux__
-		if (RUNLEVEL && strcmp(RUNLEVEL, "S") == 0) {
-			execl("/sbin/sulogin", "/sbin/sulogin", (char *) NULL);
-			eerrorx("%s: unable to exec `/sbin/sulogin': %s",
-				applet, strerror(errno));
-		}
-#endif
 		exit(EXIT_SUCCESS);
 	}
-
 #ifdef __linux__
 	run_program(SULOGIN);
 #else
@@ -354,31 +337,33 @@ static void sulogin(bool cont)
 #endif
 }
 
-_dead static void single_user(void)
+_dead static void
+single_user(void)
 {
 	rc_logger_close();
-
 	execl(SHUTDOWN, SHUTDOWN, "now", (char *) NULL);
 	eerrorx("%s: unable to exec `" SHUTDOWN "': %s",
 		 applet, strerror(errno));
 }
 
-static bool set_krunlevel(const char *level)
+static bool
+set_krunlevel(const char *level)
 {
 	FILE *fp;
 
-	if (! level ||
+	if (!level ||
 	    strcmp(level, getenv ("RC_BOOTLEVEL")) == 0 ||
 	    strcmp(level, RC_LEVEL_SINGLE) == 0 ||
 	    strcmp(level, RC_LEVEL_SYSINIT) == 0)
 	{
 		if (exists(RC_KRUNLEVEL) &&
 		    unlink(RC_KRUNLEVEL) != 0)
-			eerror("unlink `%s': %s", RC_KRUNLEVEL, strerror(errno));
+			eerror("unlink `%s': %s", RC_KRUNLEVEL,
+			       strerror(errno));
 		return false;
 	}
 
-	if (! (fp = fopen(RC_KRUNLEVEL, "w"))) {
+	if (!(fp = fopen(RC_KRUNLEVEL, "w"))) {
 		eerror("fopen `%s': %s", RC_KRUNLEVEL, strerror(errno));
 		return false;
 	}
@@ -388,37 +373,38 @@ static bool set_krunlevel(const char *level)
 	return true;
 }
 
-static int get_krunlevel(char *buffer, int buffer_len)
+static int
+get_krunlevel(char *buffer, int buffer_len)
 {
 	FILE *fp;
 	int i = 0;
 
-	if (! exists(RC_KRUNLEVEL))
+	if (!exists(RC_KRUNLEVEL))
 		return 0;
-
-	if (! (fp = fopen(RC_KRUNLEVEL, "r"))) {
+	if (!(fp = fopen(RC_KRUNLEVEL, "r"))) {
 		eerror("fopen `%s': %s", RC_KRUNLEVEL, strerror(errno));
-		return -1;
+		return 0;
 	}
 
 	if (fgets(buffer, buffer_len, fp)) {
-		i = strlen(buffer) - 1;
-		if (buffer[i] == '\n')
-			buffer[i] = 0;
+		i = strlen(buffer);
+		if (buffer[i - 1] == '\n')
+			buffer[i - 1] = 0;
 	}
-
 	fclose(fp);
 	return i;
 }
 
-static void add_pid(pid_t pid)
+static void
+add_pid(pid_t pid)
 {
 	RC_PID *p = xmalloc(sizeof(*p));
 	p->pid = pid;
 	LIST_INSERT_HEAD(&service_pids, p, entries);
 }
 
-static void remove_pid(pid_t pid)
+static void
+remove_pid(pid_t pid)
 {
 	RC_PID *p;
 
@@ -430,12 +416,19 @@ static void remove_pid(pid_t pid)
 		}
 }
 
-static void wait_for_services(void)
+static void
+wait_for_services(void)
 {
-	while (waitpid(0, 0, 0) != -1);
+	for (;;) {
+		while (waitpid(0, 0, 0) != -1)
+			;
+		if (errno != EINTR)
+			break;
+	}
 }
 
-static void handle_signal(int sig)
+static void
+handle_signal(int sig)
 {
 	int serrno = errno;
 	char signame[10] = { '\0' };
@@ -469,15 +462,15 @@ static void handle_signal(int sig)
 		break;
 
 	case SIGINT:
-		if (! signame[0])
+		if (!signame[0])
 			snprintf(signame, sizeof(signame), "SIGINT");
 		/* FALLTHROUGH */
 	case SIGTERM:
-		if (! signame[0])
+		if (!signame[0])
 			snprintf(signame, sizeof(signame), "SIGTERM");
 		/* FALLTHROUGH */
 	case SIGQUIT:
-		if (! signame[0])
+		if (!signame[0])
 			snprintf(signame, sizeof(signame), "SIGQUIT");
 		eerrorx("%s: caught %s, aborting", applet, signame);
 		/* NOTREACHED */
@@ -496,15 +489,6 @@ static void handle_signal(int sig)
 		/* Notify plugins we are aborting */
 		rc_plugin_run(RC_HOOK_ABORT, NULL);
 
-		/* Only drop into single user mode if we're booting */
-		if ((PREVLEVEL &&
-		     (strcmp(PREVLEVEL, "S") == 0 ||
-		      strcmp(PREVLEVEL, "1") == 0)) ||
-		    (RUNLEVEL &&
-		     (strcmp(RUNLEVEL, "S") == 0 ||
-		      strcmp(RUNLEVEL, "1") == 0)))
-			single_user();
-
 		exit(EXIT_FAILURE);
 		/* NOTREACHED */
 
@@ -516,222 +500,52 @@ static void handle_signal(int sig)
 	errno = serrno;
 }
 
-static void do_coldplug(void)
-{
-	size_t l;
-	DIR *dp;
-	struct dirent *d;
-	char *service;
-	RC_STRING *s;
-#ifdef BSD
-	struct ifaddrs *ifap;
-	struct ifaddrs *ifa;
-	char *p;
-#endif
-
-	errno = 0;
-	if (!rc_conf_yesno("rc_coldplug") && errno != ENOENT)
-		return;
-
-	/* We need to ensure our state dirs exist.
-	 * We should have a better call than this, but oh well. */
-	rc_deptree_update_needed();
-
-#ifdef BSD
-	if (getifaddrs(&ifap) == 0) {
-		for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
-			if (ifa->ifa_addr->sa_family != AF_LINK)
-				continue;
-
-			l = strlen("net.") + strlen(ifa->ifa_name) + 1;
-			service = xmalloc(sizeof (char) * l);
-			snprintf(service, l, "net.%s", ifa->ifa_name);
-			if (rc_service_exists(service) &&
-			    service_plugable(service))
-				rc_service_mark(service, RC_SERVICE_COLDPLUGGED);
-			free(service);
-		}
-		freeifaddrs (ifap);
-	}
-
-	/* The mice are a little more tricky.
-	 * If we coldplug anything else, we'll probably do it here. */
-	if ((dp = opendir("/dev"))) {
-		while ((d = readdir(dp))) {
-			if (strncmp(d->d_name, "psm", 3) == 0 ||
-			    strncmp(d->d_name, "ums", 3) == 0)
-			{
-				p = d->d_name + 3;
-				if (p && isdigit((unsigned char)*p)) {
-					l = strlen("moused.") + strlen(d->d_name) + 1;
-					service = xmalloc(sizeof(char) * l);
-					snprintf (service, l, "moused.%s", d->d_name);
-					if (rc_service_exists (service) &&
-					    service_plugable (service))
-						rc_service_mark (service, RC_SERVICE_COLDPLUGGED);
-					free(service);
-				}
-			}
-		}
-		closedir (dp);
-	}
-
-#else
-
-	/* udev likes to start services before we're ready when it does
-	 * its coldplugging thing. runscript knows when we're not ready so it
-	 * stores a list of coldplugged services in DEVBOOT for us to pick up
-	 * here when we are ready for them */
-	if ((dp = opendir(DEVBOOT))) {
-		while ((d = readdir(dp))) {
-			if (d->d_name[0] == '.' &&
-			    (d->d_name[1] == '\0' ||
-			     (d->d_name[1] == '.' && d->d_name[2] == '\0')))
-				continue;
-
-			if (rc_service_exists(d->d_name) &&
-			    service_plugable(d->d_name))
-				rc_service_mark(d->d_name, RC_SERVICE_COLDPLUGGED);
-
-			l = strlen(DEVBOOT "/") + strlen(d->d_name) + 1;
-			service = xmalloc(sizeof (char) * l);
-			snprintf(service, l, DEVBOOT "/%s", d->d_name);
-			if (unlink(service))
-				eerror("%s: unlink `%s': %s", applet, service,
-					strerror(errno));
-			free(service);
-		}
-		closedir(dp);
-		rmdir(DEVBOOT);
-	}
-#endif
-
-	if (rc_yesno(getenv("EINFO_QUIET")))
-		return;
-
-	/* Load our list of coldplugged services and display them */
-	einfon("Device initiated services:%s", ecolor(ECOLOR_HILITE));
-	coldplugged_services = rc_services_in_state(RC_SERVICE_COLDPLUGGED);
-	if (coldplugged_services)
-		TAILQ_FOREACH(s, coldplugged_services, entries)
-			printf(" %s", s->value);
-	printf("%s\n", ecolor(ECOLOR_NORMAL));
-}
-
-static void do_newlevel(const char *newlevel)
+static void
+do_sysinit()
 {
 	struct utsname uts;
 	const char *sys;
-#ifdef __linux__
-	char *cmd;
-#endif
 
-	if (strcmp(newlevel, RC_LEVEL_SYSINIT) == 0
-#ifndef PREFIX
-	    && RUNLEVEL &&
-	    (strcmp(RUNLEVEL, "S") == 0 ||
-	     strcmp(RUNLEVEL, "1") == 0)
-#endif
-	   )
-	{
-		/* OK, we're either in runlevel 1 or single user mode */
+	/* exec init-early.sh if it exists
+	 * This should just setup the console to use the correct
+	 * font. Maybe it should setup the keyboard too? */
+	if (exists(INITEARLYSH))
+		run_program(INITEARLYSH);
 
-		/* exec init-early.sh if it exists
-		 * This should just setup the console to use the correct
-		 * font. Maybe it should setup the keyboard too? */
-		if (exists(INITEARLYSH))
-			run_program(INITEARLYSH);
-
-		uname(&uts);
-		printf("\n   %sOpenRC %s" VERSION "%s is starting up %s",
-		       ecolor(ECOLOR_GOOD), ecolor(ECOLOR_HILITE),
-		       ecolor(ECOLOR_NORMAL), ecolor(ECOLOR_BRACKET));
+	uname(&uts);
+	printf("\n   %sOpenRC %s" VERSION "%s is starting up %s",
+	       ecolor(ECOLOR_GOOD), ecolor(ECOLOR_HILITE),
+	       ecolor(ECOLOR_NORMAL), ecolor(ECOLOR_BRACKET));
 #ifdef BRANDING
-		printf(BRANDING " (%s)", uts.machine);
+	printf(BRANDING " (%s)", uts.machine);
 #else
-		printf("%s %s (%s)",
-		       uts.sysname,
-		       uts.release,
-		       uts.machine);
+	printf("%s %s (%s)",
+	       uts.sysname,
+	       uts.release,
+	       uts.machine);
 #endif
 
-		if ((sys = rc_sys()))
-			printf(" [%s]", sys);
+	if ((sys = rc_sys()))
+		printf(" [%s]", sys);
 
-		printf("%s\n\n", ecolor(ECOLOR_NORMAL));
+	printf("%s\n\n", ecolor(ECOLOR_NORMAL));
 
-		if (! rc_yesno(getenv ("EINFO_QUIET")) &&
-		    rc_conf_yesno("rc_interactive"))
-			printf("Press %sI%s to enter interactive boot mode\n\n",
-			       ecolor(ECOLOR_GOOD), ecolor(ECOLOR_NORMAL));
+	if (!rc_yesno(getenv ("EINFO_QUIET")) &&
+	    rc_conf_yesno("rc_interactive"))
+		printf("Press %sI%s to enter interactive boot mode\n\n",
+		       ecolor(ECOLOR_GOOD), ecolor(ECOLOR_NORMAL));
 
-		setenv("RC_RUNLEVEL", newlevel, 1);
-		rc_plugin_run(RC_HOOK_RUNLEVEL_START_IN, newlevel);
-		hook_out = RC_HOOK_RUNLEVEL_START_OUT;
-		run_program(INITSH);
+	setenv("RC_RUNLEVEL", RC_LEVEL_SYSINIT, 1);
+	run_program(INITSH);
 
-#ifdef __linux__
-		/* If we requested a runlevel, save it now */
-		if ((cmd = proc_getent("rc_runlevel"))) {
-			set_krunlevel(cmd);
-			free(cmd);
-		} else if ((cmd = proc_getent("softlevel"))) {
-			set_krunlevel(cmd);
-			free(cmd);
-		} else
-		    set_krunlevel(NULL);
-#endif
-
-		/* Setup our coldplugged services now */
-		do_coldplug();
-
-		rc_plugin_run(RC_HOOK_RUNLEVEL_START_OUT, newlevel);
-		hook_out = 0;
-
-		if (want_interactive())
-			mark_interactive();
-
-		exit(EXIT_SUCCESS);
-	} else if (strcmp(newlevel, RC_LEVEL_SINGLE) == 0) {
-#ifndef PREFIX
-		if (! RUNLEVEL ||
-		    (strcmp(RUNLEVEL, "S") != 0 &&
-		     strcmp(RUNLEVEL, "1") != 0))
-		{
-			/* Remember the current runlevel for when we come back */
-			set_krunlevel(runlevel);
-			single_user();
-		}
-#endif
-	} else if (strcmp(newlevel, RC_LEVEL_REBOOT) == 0) {
-		if (! RUNLEVEL ||
-		    strcmp(RUNLEVEL, "6") != 0)
-		{
-			rc_logger_close();
-			execl(SHUTDOWN, SHUTDOWN, "-r", "now", (char *) NULL);
-			eerrorx("%s: unable to exec `" SHUTDOWN "': %s",
-				applet, strerror(errno));
-		}
-	} else if (strcmp(newlevel, RC_LEVEL_SHUTDOWN) == 0) {
-		if (! RUNLEVEL ||
-		    strcmp(RUNLEVEL, "0") != 0)
-		{
-			rc_logger_close();
-			execl(SHUTDOWN, SHUTDOWN,
-#ifdef __linux__
-			      "-h",
-#else
-			      "-p",
-#endif
-			      "now", (char *) NULL);
-			eerrorx("%s: unable to exec `" SHUTDOWN "': %s",
-				applet, strerror(errno));
-		}
-	}
+	/* init may have mounted /proc so we can now detect or real
+	 * sys */
+	if ((sys = rc_sys()))
+		setenv("RC_SYS", sys, 1);
 }
 
-
-static bool runlevel_config(const char *service, const char *level)
+static bool
+runlevel_config(const char *service, const char *level)
 {
 	char *init = rc_service_resolve(service);
 	char *conf, *dir;
@@ -746,23 +560,24 @@ static bool runlevel_config(const char *service, const char *level)
 	retval = exists(conf);
 	free(conf);
 	free(init);
-
 	return retval;
 }
 
-static void do_stop_services(const char *newlevel, bool going_down, bool parallel)
+static void
+do_stop_services(const char *newlevel, bool parallel)
 {
 	pid_t pid;
 	RC_STRING *service, *svc1, *svc2;
 	RC_STRINGLIST *deporder, *tmplist;
 	RC_SERVICE state;
-	RC_STRINGLIST *nostop = rc_stringlist_split(rc_conf_value("rc_nostop"), " ");
+	RC_STRINGLIST *nostop;
 
-	if (! types_n) {
+	if (!types_n) {
 		types_n = rc_stringlist_new();
 		rc_stringlist_add(types_n, "needsme");
 	}
 
+	nostop = rc_stringlist_split(rc_conf_value("rc_nostop"), " ");
 	TAILQ_FOREACH_REVERSE(service, stop_services, rc_stringlist, entries)
 	{
 		state = rc_service_state(service->value);
@@ -775,14 +590,6 @@ static void do_stop_services(const char *newlevel, bool going_down, bool paralle
 			continue;
 		}
 
-		/* We always stop the service when in these runlevels */
-		if (going_down || ! start_services) {
-			pid = service_stop(service->value);
-			if (pid > 0 && ! parallel)
-				rc_waitpid(pid);
-			continue;
-		}
-
 		/* If we're in the start list then don't bother stopping us */
 		svc1 = rc_stringlist_find(start_services, service->value);
 		if (svc1) {
@@ -791,17 +598,17 @@ static void do_stop_services(const char *newlevel, bool going_down, bool paralle
 				 * be stopped if we have a runlevel
 				 * configuration file for either the current
 				 * or next so we use the correct one. */
-				if (! runlevel_config(service->value, runlevel) &&
-				    ! runlevel_config(service->value, newlevel))
+				if (!runlevel_config(service->value,runlevel) &&
+				    !runlevel_config(service->value,newlevel))
 					continue;
 			}
 			else
 				continue;
 		}
 
-		/* We got this far! Or last check is to see if any any service
+		/* We got this far. Last check is to see if any any service
 		 * that going to be started depends on us */
-		if (! svc1) {
+		if (!svc1) {
 			tmplist = rc_stringlist_new();
 			rc_stringlist_add(tmplist, service->value);
 			deporder = rc_deptree_depends(deptree, types_n, tmplist,
@@ -809,7 +616,7 @@ static void do_stop_services(const char *newlevel, bool going_down, bool paralle
 						      RC_DEP_STRICT | RC_DEP_TRACE);
 			rc_stringlist_free(tmplist);
 			svc2 = NULL;
-			TAILQ_FOREACH (svc1, deporder, entries) {
+			TAILQ_FOREACH(svc1, deporder, entries) {
 				svc2 = rc_stringlist_find(start_services, svc1->value);
 				if (svc2)
 					break;
@@ -824,7 +631,7 @@ static void do_stop_services(const char *newlevel, bool going_down, bool paralle
 		pid = service_stop(service->value);
 		if (pid > 0) {
 			add_pid(pid);
-			if (! parallel) {
+			if (!parallel) {
 				rc_waitpid(pid);
 				remove_pid(pid);
 			}
@@ -834,14 +641,15 @@ static void do_stop_services(const char *newlevel, bool going_down, bool paralle
 	rc_stringlist_free(nostop);
 }
 
-static void do_start_services(bool parallel)
+static void
+do_start_services(bool parallel)
 {
 	RC_STRING *service;
 	pid_t pid;
 	bool interactive = false;
 	RC_SERVICE state;
 
-	if (! rc_yesno(getenv("EINFO_QUIET")))
+	if (!rc_yesno(getenv("EINFO_QUIET")))
 		interactive = exists(INTERACTIVE);
 
 	TAILQ_FOREACH(service, start_services, entries) {
@@ -849,7 +657,7 @@ static void do_start_services(bool parallel)
 		if (!(state & RC_SERVICE_STOPPED) || state & RC_SERVICE_FAILED)
 			continue;
 
-		if (! interactive)
+		if (!interactive)
 			interactive = want_interactive();
 
 		if (interactive) {
@@ -872,12 +680,10 @@ interactive_option:
 		}
 
 		pid = service_start(service->value);
-
 		/* Remember the pid if we're running in parallel */
 		if (pid > 0) {
 			add_pid(pid);
-
-			if (! parallel) {
+			if (!parallel) {
 				rc_waitpid(pid);
 				remove_pid(pid);
 			}
@@ -895,7 +701,8 @@ interactive_option:
 }
 
 #ifdef RC_DEBUG
-static void handle_bad_signal(int sig)
+static void
+handle_bad_signal(int sig)
 {
 	char pid[10];
 	int status;
@@ -908,12 +715,12 @@ static void handle_bad_signal(int sig)
 		case 0:
 			sprintf(pid, "%i", crashed_pid);
 			printf("\nAuto launching gdb!\n\n");
-			_exit(execlp("gdb", "gdb", "--quiet", "--pid", pid, "-ex", "bt full", NULL));
+			_exit(execlp("gdb", "gdb", "--quiet", "--pid", pid,
+				     "-ex", "bt full", NULL));
 			/* NOTREACHED */
 		default:
 			wait(&status);
 	}
-
 	_exit(1);
 	/* NOTREACHED */
 }
@@ -935,7 +742,8 @@ static const char * const longopts_help[] = {
 };
 #include "_usage.c"
 
-int main(int argc, char **argv)
+int
+main(int argc, char **argv)
 {
 	const char *bootlevel = NULL;
 	char *newlevel = NULL;
@@ -964,7 +772,7 @@ int main(int argc, char **argv)
 	applet = basename_c(argv[0]);
 	LIST_INIT(&service_pids);
 	atexit(cleanup);
-	if (! applet)
+	if (!applet)
 		eerrorx("arguments required");
 
 	if (argc > 1 && (strcmp(argv[1], "--version") == 0)) {
@@ -988,11 +796,6 @@ int main(int argc, char **argv)
 	/* Change dir to / to ensure all scripts don't use stuff in pwd */
 	chdir("/");
 
-	/* RUNLEVEL is set by sysvinit as is a magic number
-	 * RC_RUNLEVEL is set by us and is the name for this magic number */
-	RUNLEVEL = getenv("RUNLEVEL");
-	PREVLEVEL = getenv("PREVLEVEL");
-
 	/* Ensure our environment is pure
 	 * Also, add our configuration to it */
 	env_filter();
@@ -1007,7 +810,14 @@ int main(int argc, char **argv)
 		case 'o':
 			if (*optarg == '\0')
 				optarg = NULL;
-			exit(set_krunlevel(optarg) ? EXIT_SUCCESS : EXIT_FAILURE);
+			if (!rc_runlevel_exists(optarg)) {
+				eerror("runlevel `%s' does not exist", optarg);
+				exit(EXIT_FAILURE);
+			}
+			if (!set_krunlevel(optarg))
+				exit(EXIT_FAILURE);
+			einfo("Overriding next runlevel to %s", optarg);
+			exit(EXIT_SUCCESS);
 			/* NOTREACHED */
 		case 's':
 			newlevel = rc_service_resolve(optarg);
@@ -1030,6 +840,11 @@ int main(int argc, char **argv)
 	}
 
 	newlevel = argv[optind++];
+	/* For compat with old system */
+	if (newlevel) {
+		if (strcmp(newlevel, "reboot") == 0)
+			newlevel = UNCONST(RC_LEVEL_SHUTDOWN);
+	}
 
 	/* Enable logging */
 	setenv("EINFO_LOG", "rc", 1);
@@ -1051,54 +866,52 @@ int main(int argc, char **argv)
 	signal_setup(SIGUSR1, handle_signal);
 	signal_setup(SIGWINCH, handle_signal);
 
-	rc_plugin_load();
+	/* Run any special sysinit foo */
+	if (newlevel && strcmp(newlevel, RC_LEVEL_SYSINIT) == 0) {
+		do_sysinit();
+		free(runlevel);
+		runlevel = rc_runlevel_get();
+	}
 
-	/* Check we're in the runlevel requested, ie from
-	 * rc single
-	 * rc shutdown
-	 * rc reboot
-	 */
-	if (newlevel)
-		do_newlevel(newlevel);
+	rc_plugin_load();
 
 	/* Now we start handling our children */
 	signal_setup(SIGCHLD, handle_signal);
 
-	/* We should only use krunlevel if we were in single user mode
-	 * If not, we need to erase krunlevel now. */
-	if (PREVLEVEL &&
-	    (strcmp(PREVLEVEL, "1") == 0 ||
-	     strcmp(PREVLEVEL, "S") == 0 ||
-	     strcmp(PREVLEVEL, "N") == 0))
-	{
-		/* Try not to join boot and krunlevels together */
-		if (! newlevel ||
-		    strcmp(newlevel, getenv("RC_BOOTLEVEL")) != 0)
-			if (get_krunlevel(krunlevel, sizeof(krunlevel)))
-				newlevel = krunlevel;
-	} else if (! RUNLEVEL ||
-		   (strcmp(RUNLEVEL, "1") != 0 &&
-		    strcmp(RUNLEVEL, "S") != 0 &&
-		    strcmp(RUNLEVEL, "N") != 0))
-	{
-		set_krunlevel(NULL);
-	}
-
 	if (newlevel &&
-	    (strcmp(newlevel, RC_LEVEL_REBOOT) == 0 ||
-	     strcmp(newlevel, RC_LEVEL_SHUTDOWN) == 0 ||
+	    (strcmp(newlevel, RC_LEVEL_SHUTDOWN) == 0 ||
 	     strcmp(newlevel, RC_LEVEL_SINGLE) == 0))
 	{
 		going_down = true;
+		if (!exists(RC_KRUNLEVEL))
+			set_krunlevel(runlevel);
 		rc_runlevel_set(newlevel);
 		setenv("RC_RUNLEVEL", newlevel, 1);
 		setenv("RC_GOINGDOWN", "YES", 1);
+	} else {
+		/* We should not use krunevel in sysinit or the boot runlevel */
+		if (!newlevel ||
+		    strcmp(newlevel, RC_LEVEL_SYSINIT) != 0 ||
+		    strcmp(newlevel, getenv("RC_BOOTLEVEL")) != 0)
+		{
+			if (get_krunlevel(krunlevel, sizeof(krunlevel))) {
+				newlevel = krunlevel;
+				set_krunlevel(NULL);
+			}
+		}
 
+		if (newlevel &&
+		    strcmp(runlevel, newlevel) != 0 &&
+		    !rc_runlevel_exists(newlevel))
+			eerrorx("%s: is not a valid runlevel", newlevel);
+	}
+
+	if (going_down) {
 #ifdef __FreeBSD__
 		/* FIXME: we shouldn't have todo this */
-		/* For some reason, wait_for_services waits for the logger proccess
-		 * to finish as well, but only on FreeBSD. We cannot allow this so
-		 * we stop logging now. */
+		/* For some reason, wait_for_services waits for the logger
+		 * proccess to finish as well, but only on FreeBSD.
+		 * We cannot allow this so we stop logging now. */
 		rc_logger_close();
 #endif
 
@@ -1109,9 +922,9 @@ int main(int argc, char **argv)
 	hook_out = RC_HOOK_RUNLEVEL_STOP_OUT;
 
 	/* Check if runlevel is valid if we're changing */
-	if (newlevel && strcmp(runlevel, newlevel) != 0 && ! going_down) {
-		if (! rc_runlevel_exists(newlevel))
-			eerrorx ("%s: is not a valid runlevel", newlevel);
+	if (newlevel && strcmp(runlevel, newlevel) != 0 && !going_down) {
+		if (!rc_runlevel_exists(newlevel))
+			eerrorx("%s: is not a valid runlevel", newlevel);
 	}
 
 	/* Load our deptree */
@@ -1132,21 +945,11 @@ int main(int argc, char **argv)
 	 * correct order for stopping them */
 	stop_services = rc_services_in_state(RC_SERVICE_STARTED);
 	tmplist = rc_services_in_state(RC_SERVICE_INACTIVE);
-	if (tmplist) {
-		if (stop_services) {
-			TAILQ_CONCAT(stop_services, tmplist, entries);
-			free(tmplist);
-		} else
-			stop_services = tmplist;
-	}
+	TAILQ_CONCAT(stop_services, tmplist, entries);
+	free(tmplist);
 	tmplist = rc_services_in_state(RC_SERVICE_STARTING);
-	if (tmplist) {
-		if (stop_services) {
-			TAILQ_CONCAT(stop_services, tmplist, entries);
-			free(tmplist);
-		} else
-			stop_services = tmplist;
-	}
+	TAILQ_CONCAT(stop_services, tmplist, entries);
+	free(tmplist);
 	if (stop_services)
 		rc_stringlist_sort(&stop_services);
 
@@ -1162,30 +965,32 @@ int main(int argc, char **argv)
 		stop_services = tmplist;
 	}
 
-	/* Load our list of coldplugged services */
-	coldplugged_services = rc_services_in_state(RC_SERVICE_COLDPLUGGED);
-	if (strcmp(newlevel ? newlevel : runlevel, RC_LEVEL_SINGLE) != 0 &&
-	    strcmp(newlevel ? newlevel : runlevel, RC_LEVEL_SHUTDOWN) != 0 &&
-	    strcmp(newlevel ? newlevel : runlevel, RC_LEVEL_REBOOT) != 0)
+	/* Load our list of start services */
+	hotplugged_services = rc_services_in_state(RC_SERVICE_HOTPLUGGED);
+	start_services = rc_services_in_runlevel(newlevel ?
+						 newlevel : runlevel);
+	if (strcmp(newlevel ? newlevel : runlevel, RC_LEVEL_SHUTDOWN) != 0 &&
+	    strcmp(newlevel ? newlevel : runlevel, RC_LEVEL_SYSINIT) != 0)
 	{
-		/* We need to include the boot runlevel services if we're not in it */
-		start_services = rc_services_in_runlevel(bootlevel);
-		if (strcmp (newlevel ? newlevel : runlevel, bootlevel) != 0) {
-			tmplist = rc_services_in_runlevel(newlevel ? newlevel : runlevel);
-			if (tmplist) {
-				if (start_services) {
-					TAILQ_CONCAT(start_services, tmplist, entries);
-					free(tmplist);
-				} else
-					start_services = tmplist;
+		tmplist = rc_services_in_runlevel(RC_LEVEL_SYSINIT);
+		TAILQ_CONCAT(start_services, tmplist, entries);
+		free(tmplist);
+		if (strcmp(newlevel ? newlevel : runlevel,
+			   RC_LEVEL_SINGLE) != 0)
+		{
+			if (strcmp(newlevel ? newlevel : runlevel,
+				   bootlevel) != 0)
+			{
+				tmplist = rc_services_in_runlevel(bootlevel);
+				TAILQ_CONCAT(start_services, tmplist, entries);
+				free(tmplist);
+			}
+			if (hotplugged_services) {
+				TAILQ_FOREACH(service, hotplugged_services,
+					      entries)
+					rc_stringlist_addu(start_services,
+							   service->value);
 			}
-		}
-
-		if (coldplugged_services) {
-			if (!start_services)
-				start_services = rc_stringlist_new();
-			TAILQ_FOREACH(service, coldplugged_services, entries)
-				rc_stringlist_addu(start_services, service->value);
 		}
 	}
 
@@ -1193,7 +998,7 @@ int main(int argc, char **argv)
 
 	/* Now stop the services that shouldn't be running */
 	if (stop_services)
-		do_stop_services(newlevel, going_down, parallel);
+		do_stop_services(newlevel, parallel);
 
 	/* Wait for our services to finish */
 	wait_for_services();
@@ -1213,81 +1018,72 @@ int main(int argc, char **argv)
 		setenv("RC_RUNLEVEL", runlevel, 1);
 	}
 
-	/* Run the halt script if needed */
-	if (strcmp(runlevel, RC_LEVEL_SHUTDOWN) == 0 ||
-	    strcmp(runlevel, RC_LEVEL_REBOOT) == 0)
-	{
+#ifdef __linux__
+	/* We can't log beyond this point as the shutdown runlevel
+	 * will mount / readonly. */
+	if (strcmp(runlevel, RC_LEVEL_SHUTDOWN) == 0)
 		rc_logger_close();
-		execl(HALTSH, HALTSH, runlevel, (char *) NULL);
-		eerrorx("%s: unable to exec `%s': %s",
-			 applet, HALTSH, strerror(errno));
-	}
-
-	/* Single user is done now */
-	if (strcmp(runlevel, RC_LEVEL_SINGLE) == 0) {
-		if (exists(INTERACTIVE))
-			unlink(INTERACTIVE);
-		sulogin(false);
-	}
+#endif
 
 	mkdir(RC_STARTING, 0755);
 	rc_plugin_run(RC_HOOK_RUNLEVEL_START_IN, runlevel);
 	hook_out = RC_HOOK_RUNLEVEL_START_OUT;
 
-	/* Re-add our coldplugged services if they stopped */
-	if (coldplugged_services)
-		TAILQ_FOREACH(service, coldplugged_services, entries)
-			rc_service_mark(service->value, RC_SERVICE_COLDPLUGGED);
+	/* Re-add our hotplugged services if they stopped */
+	if (hotplugged_services)
+		TAILQ_FOREACH(service, hotplugged_services, entries)
+			rc_service_mark(service->value, RC_SERVICE_HOTPLUGGED);
 
 	/* Order the services to start */
 	if (start_services) {
 		rc_stringlist_sort(&start_services);
-		deporder = rc_deptree_depends(deptree, types_nua, start_services,
-				      runlevel, depoptions | RC_DEP_START);
+		deporder = rc_deptree_depends(deptree, types_nua,
+					      start_services, runlevel,
+					      depoptions | RC_DEP_START);
 		rc_stringlist_free(start_services);
 		start_services = deporder;
 	}
 
 #ifdef __linux__
 	/* mark any services skipped as started */
-	if (PREVLEVEL && strcmp(PREVLEVEL, "N") == 0) {
-		proc = p = proc_getent("noinitd");
-		if (proc) {
-			while ((token = strsep(&p, ",")))
-				rc_service_mark(token, RC_SERVICE_STARTED);
-			free(proc);
-		}
+	proc = p = proc_getent("noinitd");
+	if (proc) {
+		while ((token = strsep(&p, ",")))
+			rc_service_mark(token, RC_SERVICE_STARTED);
+		free(proc);
 	}
 #endif
 
 	if (start_services) {
 		do_start_services(parallel);
+		/* FIXME: If we skip the boot runlevel and go straight
+		 * to default from sysinit, we should now re-evaluate our
+		 * start services + hotplugged services and call
+		 * do_start_services a second time. */
 
 		/* Wait for our services to finish */
 		wait_for_services();
 	}
 
-	rc_plugin_run(RC_HOOK_RUNLEVEL_START_OUT, runlevel);
-	hook_out = 0;
-
 #ifdef __linux__
 	/* mark any services skipped as stopped */
-	if (PREVLEVEL && strcmp(PREVLEVEL, "N") == 0) {
-		proc = p = proc_getent("noinitd");
-		if (proc) {
-			while ((token = strsep(&p, ",")))
-				rc_service_mark(token, RC_SERVICE_STOPPED);
-			free(proc);
-		}
+	proc = p = proc_getent("noinitd");
+	if (proc) {
+		while ((token = strsep(&p, ",")))
+			rc_service_mark(token, RC_SERVICE_STOPPED);
+		free(proc);
 	}
 #endif
 
+	rc_plugin_run(RC_HOOK_RUNLEVEL_START_OUT, runlevel);
+	hook_out = 0;
+
 	/* If we're in the boot runlevel and we regenerated our dependencies
 	 * we need to delete them so that they are regenerated again in the
-	 * default runlevel as they may depend on things that are now available */
+	 * default runlevel as they may depend on things that are now
+	 * available */
 	if (regen && strcmp(runlevel, bootlevel) == 0)
 		unlink(RC_DEPTREE_CACHE);
 
 	return EXIT_SUCCESS;
 }
-
