@@ -29,19 +29,21 @@
  * SUCH DAMAGE.
  */
 
-#include <sys/select.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
 #include <sys/file.h>
 #include <sys/param.h>
 #include <sys/stat.h>
 
+#include <ctype.h>
 #include <dlfcn.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <fnmatch.h>
 #include <getopt.h>
 #include <libgen.h>
 #include <limits.h>
+#include <poll.h>
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -108,9 +110,8 @@ static RC_STRINGLIST *types_mua = NULL;
 static void (*selinux_run_init_old)(void);
 static void (*selinux_run_init_new)(int argc, char **argv);
 
-static void setup_selinux(int argc, char **argv);
-
-static void setup_selinux(int argc, char **argv)
+static void
+setup_selinux(int argc, char **argv)
 {
 	void *lib_handle = NULL;
 
@@ -141,7 +142,8 @@ static void setup_selinux(int argc, char **argv)
 }
 #endif
 
-static void handle_signal(int sig)
+static void
+handle_signal(int sig)
 {
 	int serrno = errno;
 	char signame[10] = { '\0' };
@@ -155,7 +157,8 @@ static void handle_signal(int sig)
 	case SIGCHLD:
 		if (signal_pipe[1] > -1) {
 			if (write(signal_pipe[1], &sig, sizeof(sig)) == -1)
-				eerror("%s: send: %s", service, strerror(errno));
+				eerror("%s: send: %s",
+				       service, strerror(errno));
 		} else
 			rc_waitpid(-1);
 		break;
@@ -168,15 +171,15 @@ static void handle_signal(int sig)
 		break;
 
 	case SIGINT:
-		if (! signame[0])
+		if (!signame[0])
 			snprintf(signame, sizeof(signame), "SIGINT");
 		/* FALLTHROUGH */
 	case SIGTERM:
-		if (! signame[0])
+		if (!signame[0])
 			snprintf(signame, sizeof(signame), "SIGTERM");
 		/* FALLTHROUGH */
 	case SIGQUIT:
-		if (! signame[0])
+		if (!signame[0])
 			snprintf(signame, sizeof(signame), "SIGQUIT");
 		/* Send the signal to our children too */
 		if (service_pid > 0)
@@ -192,18 +195,17 @@ static void handle_signal(int sig)
 	errno = serrno;
 }
 
-static time_t get_mtime(const char *pathname, bool follow_link)
+static time_t
+get_mtime(const char *pathname, bool follow_link)
 {
 	struct stat buf;
 	int retval;
 
-	if (! pathname)
+	if (!pathname)
 		return 0;
-
 	retval = follow_link ? stat(pathname, &buf) : lstat(pathname, &buf);
-	if (! retval)
+	if (!retval)
 		return buf.st_mtime;
-
 	errno = 0;
 	return 0;
 }
@@ -211,7 +213,8 @@ static time_t get_mtime(const char *pathname, bool follow_link)
 static const char *const tests[] = {
 	"starting", "started", "stopping", "inactive", "wasinactive", NULL
 };
-static bool in_control()
+static bool
+in_control()
 {
 	char file[PATH_MAX];
 	time_t m;
@@ -221,17 +224,18 @@ static bool in_control()
 	if (sighup)
 		return false;
 
-	if (! *mtime_test || ! exists(mtime_test))
+	if (!*mtime_test || !exists(mtime_test))
 		return false;
 
 	if (rc_service_state(applet) & RC_SERVICE_STOPPED)
 		return false;
 
-	if (! (mtime = get_mtime(mtime_test, false)))
+	if (!(mtime = get_mtime(mtime_test, false)))
 		return false;
 
 	while (tests[i]) {
-		snprintf(file, sizeof(file), RC_SVCDIR "/%s/%s", tests[i], applet);
+		snprintf(file, sizeof(file), RC_SVCDIR "/%s/%s",
+			 tests[i], applet);
 		if (exists(file)) {
 			m = get_mtime(file, false);
 			if (mtime < m && m != 0)
@@ -243,20 +247,23 @@ static bool in_control()
 	return true;
 }
 
-static void uncoldplug()
+static void
+unhotplug()
 {
 	char file[PATH_MAX];
 
-	snprintf(file, sizeof(file), RC_SVCDIR "/coldplugged/%s", applet);
+	snprintf(file, sizeof(file), RC_SVCDIR "/hotplugged/%s", applet);
 	if (exists(file) && unlink(file) != 0)
 		eerror("%s: unlink `%s': %s", applet, file, strerror(errno));
 }
 
-static void start_services(RC_STRINGLIST *list) {
+static void
+start_services(RC_STRINGLIST *list)
+{
 	RC_STRING *svc;
 	RC_SERVICE state = rc_service_state (service);
 
-	if (! list)
+	if (!list)
 		return;
 
 	if (state & RC_SERVICE_INACTIVE ||
@@ -265,28 +272,30 @@ static void start_services(RC_STRINGLIST *list) {
 	    state & RC_SERVICE_STARTED)
 	{
 		TAILQ_FOREACH(svc, list, entries) {
-			if (rc_service_state(svc->value) & RC_SERVICE_STOPPED) {
-				if (state & RC_SERVICE_INACTIVE ||
-				    state & RC_SERVICE_WASINACTIVE)
-				{
-					rc_service_schedule_start(service, svc->value);
-					ewarn("WARNING: %s is scheduled to started"
-					      " when %s has started",
-					       svc->value, applet);
-				} else
-					service_start(svc->value);
-			}
+			if (!(rc_service_state(svc->value) &
+			      RC_SERVICE_STOPPED))
+				continue;
+			if (state & RC_SERVICE_INACTIVE ||
+			    state & RC_SERVICE_WASINACTIVE)
+			{
+				rc_service_schedule_start(service,
+							  svc->value);
+				ewarn("WARNING: %s is scheduled to started"
+				      " when %s has started",
+				       svc->value, applet);
+			} else
+				service_start(svc->value);
 		}
 	}
 }
 
-static void restore_state(void)
+static void
+restore_state(void)
 {
 	RC_SERVICE state;
 
-	if (rc_in_plugin || ! in_control())
+	if (rc_in_plugin || !in_control())
 		return;
-
 	state = rc_service_state(applet);
 	if (state & RC_SERVICE_STOPPING) {
 		if (state & RC_SERVICE_WASINACTIVE)
@@ -310,50 +319,56 @@ static void restore_state(void)
 	}
 }
 
-static void cleanup(void)
+static void
+cleanup(void)
 {
 	restore_state();
 
-	if (! rc_in_plugin) {
+	if (!rc_in_plugin) {
 		if (hook_out) {
 			rc_plugin_run(hook_out, applet);
 			if (hook_out == RC_HOOK_SERVICE_START_DONE)
-				rc_plugin_run(RC_HOOK_SERVICE_START_OUT, applet);
+				rc_plugin_run(RC_HOOK_SERVICE_START_OUT,
+					      applet);
 			else if (hook_out == RC_HOOK_SERVICE_STOP_DONE)
-				rc_plugin_run(RC_HOOK_SERVICE_STOP_OUT, applet);
+				rc_plugin_run(RC_HOOK_SERVICE_STOP_OUT,
+					      applet);
 		}
 
 		if (restart_services)
 			start_services(restart_services);
 	}
 
+	rc_plugin_unload();
+
+#ifdef DEBUG_MEMORY
 	rc_stringlist_free(types_b);
 	rc_stringlist_free(types_n);
 	rc_stringlist_free(types_nu);
 	rc_stringlist_free(types_nua);
 	rc_stringlist_free(types_m);
 	rc_stringlist_free(types_mua);
-	
-	rc_plugin_unload();
 	rc_deptree_free(deptree);
-
 	rc_stringlist_free(restart_services);
 	rc_stringlist_free(need_services);
 	rc_stringlist_free(use_services);
 	rc_stringlist_free(services);
 	rc_stringlist_free(applet_list);
 	rc_stringlist_free(tmplist);
-	free (ibsave);
+	free(ibsave);
 	free(service);
 	free(prefix);
 	free(runlevel);
+#endif
 
-	if (*mtime_test && ! rc_in_plugin)
+	if (*mtime_test && !rc_in_plugin)
 		unlink(mtime_test);
 }
 
-static int write_prefix(const char *buffer, size_t bytes, bool *prefixed) {
-	unsigned int i;
+static int
+write_prefix(const char *buffer, size_t bytes, bool *prefixed)
+{
+	size_t i, j;
 	const char *ec = ecolor(ECOLOR_HILITE);
 	const char *ec_normal = ecolor(ECOLOR_NORMAL);
 	ssize_t ret = 0;
@@ -369,11 +384,17 @@ static int write_prefix(const char *buffer, size_t bytes, bool *prefixed) {
 	}
 
 	for (i = 0; i < bytes; i++) {
-		/* We don't prefix escape codes, like eend */
-		if (buffer[i] == '\033')
-			*prefixed = true;
+		/* We don't prefix eend calls (cursor up) */
+		if (buffer[i] == '\033' && !*prefixed) {
+			for (j = i + 1; j < bytes; j++) {
+				if (buffer[j] == 'A')
+					*prefixed = true;
+				if (isalpha((unsigned int)buffer[j]))
+					break;
+			}
+		}
 
-		if (! *prefixed) {
+		if (!*prefixed) {
 			ret += write(fd, ec, strlen(ec));
 			ret += write(fd, prefix, strlen(prefix));
 			ret += write(fd, ec_normal, strlen(ec_normal));
@@ -388,11 +409,11 @@ static int write_prefix(const char *buffer, size_t bytes, bool *prefixed) {
 
 	/* Release the lock */
 	close(lock_fd);
-
 	return ret;
 }
 
-static bool svc_exec(const char *arg1, const char *arg2)
+static bool
+svc_exec(const char *arg1, const char *arg2)
 {
 	bool execok;
 	int fdout = fileno(stdout);
@@ -400,12 +421,11 @@ static bool svc_exec(const char *arg1, const char *arg2)
 	struct winsize ws;
 	int i;
 	int flags = 0;
-	fd_set rset;
+	struct pollfd fd[2];
 	int s;
 	char *buffer;
 	size_t bytes;
 	bool prefixed = false;
-	int selfd;
 	int slave_tty;
 
 	/* Setup our signal pipe */
@@ -429,7 +449,6 @@ static bool svc_exec(const char *arg1, const char *arg2)
 		/* If the below call fails due to not enough ptys then we don't
 		 * prefix the output, but we still work */
 		openpty(&master_tty, &slave_tty, NULL, &tt, &ws);
-
 		if (master_tty >= 0 &&
 		    (flags = fcntl(master_tty, F_GETFD, 0)) == 0)
 		     fcntl(master_tty, F_SETFD, flags | FD_CLOEXEC);
@@ -444,18 +463,20 @@ static bool svc_exec(const char *arg1, const char *arg2)
 		eerrorx("%s: fork: %s", service, strerror(errno));
 	if (service_pid == 0) {
 		if (slave_tty >= 0) {
-			dup2(slave_tty, 1);
-			dup2(slave_tty, 2);
+			dup2(slave_tty, STDOUT_FILENO);
+			dup2(slave_tty, STDERR_FILENO);
 		}
 
 		if (exists(RC_SVCDIR "/runscript.sh")) {
-			execl(RC_SVCDIR "/runscript.sh", RC_SVCDIR "/runscript.sh",
+			execl(RC_SVCDIR "/runscript.sh",
+			      RC_SVCDIR "/runscript.sh",
 			      service, arg1, arg2, (char *) NULL);
 			eerror("%s: exec `" RC_SVCDIR "/runscript.sh': %s",
 				service, strerror(errno));
 			_exit(EXIT_FAILURE);
 		} else {
-			execl(RC_LIBDIR "/sh/runscript.sh", RC_LIBDIR "/sh/runscript.sh",
+			execl(RC_LIBDIR "/sh/runscript.sh",
+			      RC_LIBDIR "/sh/runscript.sh",
 			      service, arg1, arg2, (char *) NULL);
 			eerror("%s: exec `" RC_LIBDIR "/sh/runscript.sh': %s",
 			       service, strerror(errno));
@@ -463,29 +484,33 @@ static bool svc_exec(const char *arg1, const char *arg2)
 		}
 	}
 
-	selfd = MAX(master_tty, signal_pipe[0]) + 1;
 	buffer = xmalloc(sizeof(char) * BUFSIZ);
-	for (;;) {
-		FD_ZERO(&rset);
-		FD_SET(signal_pipe[0], &rset);
-		if (master_tty >= 0)
-			FD_SET(master_tty, &rset);
+	fd[0].fd = signal_pipe[0];
+	fd[0].events = fd[1].events = POLLIN;
+	fd[0].revents = fd[1].revents = 0;
+	if (master_tty >= 0) {
+		fd[1].fd = master_tty;
+		fd[1].events = POLLIN;
+		fd[1].revents = 0;
+	}
 
-		if ((s = select(selfd, &rset, NULL, NULL, NULL)) == -1) {
+	for (;;) {
+		if ((s = poll(fd, master_tty >= 0 ? 2 : 1, -1)) == -1) {
 			if (errno != EINTR) {
-				eerror("%s: select: %s", service, strerror(errno));
+				eerror("%s: poll: %s",
+					service, strerror(errno));
 				break;
 			}
 		}
 
 		if (s > 0) {
-			if (master_tty >= 0 && FD_ISSET(master_tty, &rset)) {
+			if (fd[1].revents & (POLLIN | POLLHUP)) {
 				bytes = read(master_tty, buffer, BUFSIZ);
 				write_prefix(buffer, bytes, &prefixed);
 			}
 
 			/* Only SIGCHLD signals come down this pipe */
-			if (FD_ISSET(signal_pipe[0], &rset))
+			if (fd[0].revents & (POLLIN | POLLHUP))
 				break;
 		}
 	}
@@ -503,12 +528,16 @@ static bool svc_exec(const char *arg1, const char *arg2)
 	}
 
 	execok = rc_waitpid(service_pid) == 0 ? true : false;
+	if (!execok && errno == ECHILD)
+		/* killall5 -9 could cause this */
+		execok = true;
 	service_pid = 0;
 
 	return execok;
 }
 
-static bool svc_wait(const char *svc)
+static bool
+svc_wait(const char *svc)
 {
 	char fifo[PATH_MAX];
 	struct timespec ts;
@@ -524,12 +553,13 @@ static bool svc_wait(const char *svc)
 		forever = true;
 	rc_stringlist_free(keywords);
 
-	snprintf(fifo, sizeof(fifo), RC_SVCDIR "/exclusive/%s", basename_c(svc));
+	snprintf(fifo, sizeof(fifo), RC_SVCDIR "/exclusive/%s",
+		 basename_c(svc));
 	ts.tv_sec = 0;
 	ts.tv_nsec = WAIT_INTERVAL;
 
 	while (nloops) {
-		if (! exists(fifo)) {
+		if (!exists(fifo)) {
 			retval = true;
 			break;
 		}
@@ -539,24 +569,26 @@ static bool svc_wait(const char *svc)
 				break;
 		}
 
-		if (! forever)
+		if (!forever) {
 			nloops --;
 
-		if (--sloops == 0) {
-			ewarn("%s: waiting for %s", applet, svc);
-			sloops = (ONE_SECOND / WAIT_INTERVAL) * 5;
+			if (--sloops == 0) {
+				ewarn("%s: waiting for %s", applet, svc);
+				sloops = (ONE_SECOND / WAIT_INTERVAL) * 5;
+			}
 		}
 	}
 
-	if (! exists(fifo))
+	if (!exists(fifo))
 		retval = true;
 	return retval;
 }
 
-static RC_SERVICE svc_status(void)
+static RC_SERVICE
+svc_status(void)
 {
 	char status[10];
-	int (*e) (const char *fmt, ...) __EINFO_PRINTF = einfo;
+	int (*e) (const char *fmt, ...) EINFO_PRINTF(1, 2) = einfo;
 	RC_SERVICE state = rc_service_state(service);
 
 	if (state & RC_SERVICE_STOPPING) {
@@ -585,19 +617,21 @@ static RC_SERVICE svc_status(void)
 	return state;
 }
 
-static void make_exclusive(void)
+static void
+make_exclusive(void)
 {
 	/* We create a fifo so that other services can wait until we complete */
-	if (! *exclusive)
-		snprintf(exclusive, sizeof(exclusive), RC_SVCDIR "/exclusive/%s",
-			 applet);
+	if (!*exclusive)
+		snprintf(exclusive, sizeof(exclusive),
+			 RC_SVCDIR "/exclusive/%s", applet);
 
 	if (mkfifo(exclusive, 0600) != 0 && errno != EEXIST &&
 	    (errno != EACCES || geteuid () == 0))
 		eerrorx ("%s: unable to create fifo `%s': %s",
 			 applet, exclusive, strerror(errno));
 
-	snprintf(mtime_test, sizeof(mtime_test), RC_SVCDIR "/exclusive/%s.%d", applet, getpid());
+	snprintf(mtime_test, sizeof(mtime_test),
+		 RC_SVCDIR "/exclusive/%s.%d", applet, getpid());
 
 	if (exists(mtime_test) && unlink(mtime_test) != 0) {
 		eerror("%s: unlink `%s': %s",
@@ -613,28 +647,28 @@ static void make_exclusive(void)
 	}
 }
 
-static void unlink_mtime_test(void)
+static void
+unlink_mtime_test(void)
 {
 	if (unlink(mtime_test) != 0)
-		eerror("%s: unlink `%s': %s", applet, mtime_test, strerror(errno));
+		eerror("%s: unlink `%s': %s",
+		       applet, mtime_test, strerror(errno));
 	*mtime_test = '\0';
 }
 
-static void get_started_services(void)
+static void
+get_started_services(void)
 {
 	RC_STRINGLIST *tmp = rc_services_in_state(RC_SERVICE_INACTIVE);
+
 	rc_stringlist_free(restart_services);
 	restart_services = rc_services_in_state(RC_SERVICE_STARTED);
-	if (tmp) {
-		if (restart_services) {
-			TAILQ_CONCAT(restart_services, tmp, entries);
-			free(tmp);
-		} else
-			restart_services = tmp;
-	}
+	TAILQ_CONCAT(restart_services, tmp, entries);
+	free(tmp);
 }
 
-static void setup_types(void)
+static void
+setup_types(void)
 {
 	types_b = rc_stringlist_new();
 	rc_stringlist_add(types_b, "broken");
@@ -660,7 +694,8 @@ static void setup_types(void)
 	rc_stringlist_add(types_mua, "beforeme");
 }
 
-static void svc_start(bool deps)
+static void
+svc_start(bool deps)
 {
 	bool started;
 	bool background = false;
@@ -681,6 +716,10 @@ static void svc_start(bool deps)
 		    ! state & RC_SERVICE_STOPPED)
 			exit(EXIT_FAILURE);
 		background = true;
+		rc_service_mark(service, RC_SERVICE_HOTPLUGGED);
+		if (rc_runlevel_starting())
+			ewarnx("WARNING: %s will be started when the runlevel"
+			       " has finished.", applet);
 	}
 
 	if (state & RC_SERVICE_STARTED) {
@@ -691,16 +730,16 @@ static void svc_start(bool deps)
 	else if (state & RC_SERVICE_STOPPING)
 		ewarnx("WARNING: %s is stopping", applet);
 	else if (state & RC_SERVICE_INACTIVE && ! background)
-		ewarnx("WARNING: %s has already started, but is inactive", applet);
+		ewarnx("WARNING: %s has already started, but is inactive",
+		       applet);
 
-	if (! rc_service_mark(service, RC_SERVICE_STARTING)) {
+	if (!rc_service_mark(service, RC_SERVICE_STARTING)) {
 		if (errno == EACCES)
 			eerrorx("%s: superuser access required", applet);
 		eerrorx("ERROR: %s has been started by something else", applet);
 	}
 
 	make_exclusive();
-
 	hook_out = RC_HOOK_SERVICE_START_OUT;
 	rc_plugin_run(RC_HOOK_SERVICE_START_IN, applet);
 
@@ -709,15 +748,14 @@ static void svc_start(bool deps)
 		depoptions |= RC_DEP_STRICT;
 
 	if (deps) {
-		if (! deptree && ((deptree = _rc_deptree_load(NULL)) == NULL))
+		if (!deptree && ((deptree = _rc_deptree_load(NULL)) == NULL))
 			eerrorx("failed to load deptree");
-
-		if (! types_b)
+		if (!types_b)
 			setup_types();
 
 		services = rc_deptree_depends(deptree, types_b, applet_list,
 				runlevel, 0);
-		if (services && TAILQ_FIRST(services)) {
+		if (TAILQ_FIRST(services)) {
 			eerrorn("ERROR: `%s' needs ", applet);
 			first = true;
 			TAILQ_FOREACH(svc, services, entries) {
@@ -732,12 +770,14 @@ static void svc_start(bool deps)
 		rc_stringlist_free(services);
 		services = NULL;
 
-		need_services = rc_deptree_depends(deptree, types_n, applet_list,
-				runlevel, depoptions);
-		use_services = rc_deptree_depends(deptree, types_nu, applet_list,
-				runlevel, depoptions);
+		need_services = rc_deptree_depends(deptree, types_n,
+						   applet_list,	runlevel,
+						   depoptions);
+		use_services = rc_deptree_depends(deptree, types_nu,
+						  applet_list, runlevel,
+						  depoptions);
 
-		if (! rc_runlevel_starting() && use_services)
+		if (!rc_runlevel_starting()) {
 			TAILQ_FOREACH(svc, use_services, entries) {
 				state = rc_service_state(svc->value);
 				/* Don't stop failed services again.
@@ -752,90 +792,89 @@ static void svc_start(bool deps)
 						rc_waitpid(pid);
 				}
 			}
+		}
 
 		/* Now wait for them to start */
 		services = rc_deptree_depends(deptree, types_nua, applet_list,
 					      runlevel, depoptions);
+		/* We use tmplist to hold our scheduled by list */
+		tmplist = rc_stringlist_new();
+		TAILQ_FOREACH(svc, services, entries) {
+			state = rc_service_state(svc->value);
+			if (state & RC_SERVICE_STARTED)
+				continue;
 
-		if (services) {
-			/* We use tmplist to hold our scheduled by list */
-			tmplist = NULL;
-			TAILQ_FOREACH(svc, services, entries) {
-				state = rc_service_state(svc->value);
-				if (state & RC_SERVICE_STARTED)
+			/* Don't wait for services which went inactive but are
+			 * now in starting state which we are after */
+			if (state & RC_SERVICE_STARTING &&
+			    state & RC_SERVICE_WASINACTIVE)
+			{
+				if (!rc_stringlist_find(need_services,
+							svc->value) &&
+				    !rc_stringlist_find(use_services,
+							svc->value))
 					continue;
+			}
 
-				/* Don't wait for services which went inactive but are now in
-				 * starting state which we are after */
-				if (state & RC_SERVICE_STARTING &&
+			if (!svc_wait(svc->value))
+				eerror("%s: timed out waiting for %s",
+				       applet, svc->value);
+			state = rc_service_state(svc->value);
+			if (state & RC_SERVICE_STARTED)
+				continue;
+			if (rc_stringlist_find(need_services, svc->value)) {
+				if (state & RC_SERVICE_INACTIVE ||
 				    state & RC_SERVICE_WASINACTIVE)
 				{
-					if (!rc_stringlist_find(need_services, svc->value) &&
-					    !rc_stringlist_find(use_services, svc->value))
-						continue;
-				}
-
-				if (!svc_wait(svc->value))
-					eerror ("%s: timed out waiting for %s",
+					rc_stringlist_add(tmplist, svc->value);
+				} else if (!TAILQ_FIRST(tmplist))
+					eerrorx("ERROR: cannot start %s as"
+						" %s would not start",
 						applet, svc->value);
-				state = rc_service_state(svc->value);
-				if (state & RC_SERVICE_STARTED)
-					continue;
-				if (rc_stringlist_find(need_services, svc->value)) {
-					if (state & RC_SERVICE_INACTIVE ||
-					    state & RC_SERVICE_WASINACTIVE)
-					{
-						if (! tmplist)
-							tmplist = rc_stringlist_new();
-						rc_stringlist_add(tmplist, svc->value);
-					} else if (!tmplist)
-						eerrorx("ERROR: cannot start %s as"
-							" %s would not start",
-							applet, svc->value);
-				}
 			}
+		}
 
-			if (tmplist && TAILQ_FIRST(tmplist)) {
-				/* Set the state now, then unlink our exclusive so that
-				   our scheduled list is preserved */
-				rc_service_mark(service, RC_SERVICE_STOPPED);
-				unlink_mtime_test();
+		if (TAILQ_FIRST(tmplist)) {
+			/* Set the state now, then unlink our exclusive so that
+			   our scheduled list is preserved */
+			rc_service_mark(service, RC_SERVICE_STOPPED);
+			unlink_mtime_test();
 
+			rc_stringlist_free(use_services);
+			use_services = NULL;
+			len = 0;
+			n = 0;
+			TAILQ_FOREACH(svc, tmplist, entries) {
+				rc_service_schedule_start(svc->value, service);
+				use_services = rc_deptree_depend(deptree, 
+								 "iprovide",
+								 svc->value);
+				TAILQ_FOREACH(svc2, use_services, entries)
+					rc_service_schedule_start(svc2->value,
+								  service);
 				rc_stringlist_free(use_services);
 				use_services = NULL;
-				len = 0;
-				n = 0;
-				TAILQ_FOREACH(svc, tmplist, entries) {
-					rc_service_schedule_start(svc->value, service);
-					use_services = rc_deptree_depend(deptree, "iprovide",
-							svc->value);
-					if (use_services) {
-						TAILQ_FOREACH (svc2, use_services, entries)
-							rc_service_schedule_start(svc2->value, service);
-						rc_stringlist_free(use_services);
-						use_services = NULL;
-					}
-					len += strlen(svc->value) + 2;
-					n++;
-				}
-
-				len += 5;
-				tmp = p = xmalloc(sizeof(char) * len);
-				TAILQ_FOREACH(svc, tmplist, entries) {
-					if (p != tmp)
-						p += snprintf(p, len, ", ");
-					p += snprintf(p, len - (p - tmp), "%s", svc->value);
-				}
-				rc_stringlist_free(tmplist);
-				tmplist = NULL;
-				ewarnx("WARNING: %s is scheduled to start when %s has started",
-						applet, tmp);
-				free(tmp);
+				len += strlen(svc->value) + 2;
+				n++;
 			}
 
-			rc_stringlist_free(services);
-			services = NULL;
+			len += 5;
+			tmp = p = xmalloc(sizeof(char) * len);
+			TAILQ_FOREACH(svc, tmplist, entries) {
+				if (p != tmp)
+					p += snprintf(p, len, ", ");
+				p += snprintf(p, len - (p - tmp),
+					      "%s", svc->value);
+			}
+			rc_stringlist_free(tmplist);
+			tmplist = NULL;
+			ewarnx("WARNING: %s is scheduled to start when "
+			       "%s has started", applet, tmp);
+			free(tmp);
 		}
+
+		rc_stringlist_free(services);
+		services = NULL;
 	}
 
 	if (ibsave)
@@ -847,13 +886,15 @@ static void svc_start(bool deps)
 		unsetenv("IN_BACKGROUND");
 
 	if (in_control()) {
-		if (! started)
+		if (!started)
 			eerrorx("ERROR: %s failed to start", applet);
 	} else {
 		if (rc_service_state(service) & RC_SERVICE_INACTIVE)
-			ewarnx("WARNING: %s has started, but is inactive", applet);
+			ewarnx("WARNING: %s has started, but is inactive",
+			       applet);
 		else
-			ewarnx("WARNING: %s not under our control, aborting", applet);
+			ewarnx("WARNING: %s not under our control, aborting",
+			       applet);
 	}
 
 	rc_service_mark(service, RC_SERVICE_STARTED);
@@ -864,28 +905,23 @@ static void svc_start(bool deps)
 
 	/* Now start any scheduled services */
 	services = rc_services_scheduled(service);
-	if (services) {
-		TAILQ_FOREACH(svc, services, entries)
-			if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
-				service_start(svc->value);
-		rc_stringlist_free(services);
-		services = NULL;
-	}
+	TAILQ_FOREACH(svc, services, entries)
+		if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
+			service_start(svc->value);
+	rc_stringlist_free(services);
+	services = NULL;
 
 	/* Do the same for any services we provide */
 	if (deptree) {
 		tmplist = rc_deptree_depend(deptree, "iprovide", applet);
-		if (tmplist) {
-			TAILQ_FOREACH(svc, tmplist, entries) {
-				services = rc_services_scheduled(svc->value);
-				if (! services)
-					continue;
-				TAILQ_FOREACH(svc2, services, entries)
-					if (rc_service_state(svc2->value) & RC_SERVICE_STOPPED)
-						service_start(svc2->value);
-				rc_stringlist_free(services);
-				services = NULL;
-			}
+		TAILQ_FOREACH(svc, tmplist, entries) {
+			services = rc_services_scheduled(svc->value);
+			TAILQ_FOREACH(svc2, services, entries)
+				if (rc_service_state(svc2->value) &
+				    RC_SERVICE_STOPPED)
+					service_start(svc2->value);
+			rc_stringlist_free(services);
+			services = NULL;
 		}
 		rc_stringlist_free(tmplist);
 		tmplist = NULL;
@@ -895,20 +931,21 @@ static void svc_start(bool deps)
 	rc_plugin_run(RC_HOOK_SERVICE_START_OUT, applet);
 }
 
-static void svc_stop(bool deps)
+static void
+svc_stop(bool deps)
 {
 	bool stopped;
 	RC_SERVICE state = rc_service_state(service);
 	int depoptions = RC_DEP_TRACE;
 	RC_STRING *svc;
 
-	if (rc_runlevel_stopping() &&
-	    state & RC_SERVICE_FAILED)
-		exit (EXIT_FAILURE);
+
+	if (rc_runlevel_stopping() && state & RC_SERVICE_FAILED)
+		exit(EXIT_FAILURE);
 
 	if (rc_yesno(getenv("IN_HOTPLUG")) || in_background)
-		if (! (state & RC_SERVICE_STARTED) &&
-		    ! (state & RC_SERVICE_INACTIVE))
+		if (!(state & RC_SERVICE_STARTED) &&
+		    !(state & RC_SERVICE_INACTIVE))
 			exit (EXIT_FAILURE);
 
 	if (state & RC_SERVICE_STOPPED) {
@@ -917,7 +954,7 @@ static void svc_stop(bool deps)
 	} else if (state & RC_SERVICE_STOPPING)
 		ewarnx("WARNING: %s is already stopping", applet);
 
-	if (! rc_service_mark(service, RC_SERVICE_STOPPING)) {
+	if (!rc_service_mark(service, RC_SERVICE_STOPPING)) {
 		if (errno == EACCES)
 			eerrorx("%s: superuser access required", applet);
 		eerrorx("ERROR: %s has been stopped by something else", applet);
@@ -928,90 +965,85 @@ static void svc_stop(bool deps)
 	hook_out = RC_HOOK_SERVICE_STOP_OUT;
 	rc_plugin_run(RC_HOOK_SERVICE_STOP_IN, applet);
 
-	if (! rc_runlevel_stopping() &&
+	if (!rc_runlevel_stopping() &&
 	    rc_service_in_runlevel(service, RC_LEVEL_BOOT))
 		ewarn ("WARNING: you are stopping a boot service");
 
-	if (deps && ! (state & RC_SERVICE_WASINACTIVE)) {
+	if (deps && !(state & RC_SERVICE_WASINACTIVE)) {
 		errno = 0;
 		if (rc_conf_yesno("rc_depend_strict") || errno == ENOENT)
 			depoptions |= RC_DEP_STRICT;
 
-		if (! deptree && ((deptree = _rc_deptree_load(NULL)) == NULL))
-			eerrorx ("failed to load deptree");
+		if (!deptree && ((deptree = _rc_deptree_load(NULL)) == NULL))
+			eerrorx("failed to load deptree");
 
-		if (! types_m)
+		if (!types_m)
 			setup_types();
 
 		services = rc_deptree_depends(deptree, types_m, applet_list,
 					      runlevel, depoptions);
-		if (services) {
-			TAILQ_FOREACH_REVERSE(svc, services, rc_stringlist, entries) {
+		tmplist = rc_stringlist_new();
+		TAILQ_FOREACH_REVERSE(svc, services, rc_stringlist, entries) {
+			state = rc_service_state(svc->value);
+			/* Don't stop failed services again.
+			 * If you remove this check, ensure that the
+			 * exclusive file isn't created. */
+			if (state & RC_SERVICE_FAILED &&
+			    rc_runlevel_stopping())
+				continue;
+			if (state & RC_SERVICE_STARTED ||
+			    state & RC_SERVICE_INACTIVE)
+			{
+				svc_wait(svc->value);
 				state = rc_service_state(svc->value);
-				/* Don't stop failed services again.
-				 * If you remove this check, ensure that the
-				 * exclusive file isn't created. */
-				if (state & RC_SERVICE_FAILED &&
-				    rc_runlevel_stopping())
-					continue;
 				if (state & RC_SERVICE_STARTED ||
 				    state & RC_SERVICE_INACTIVE)
 				{
-					svc_wait(svc->value);
-					state = rc_service_state(svc->value);
-					if (state & RC_SERVICE_STARTED ||
-					    state & RC_SERVICE_INACTIVE)
-					{
-						pid_t pid = service_stop(svc->value);
-						if (! rc_conf_yesno("rc_parallel"))
-							rc_waitpid(pid);
-						if (! tmplist)
-							tmplist = rc_stringlist_new();
-						rc_stringlist_add(tmplist, svc->value);
-					}
+					pid_t pid = service_stop(svc->value);
+					if (!rc_conf_yesno("rc_parallel"))
+						rc_waitpid(pid);
+					rc_stringlist_add(tmplist, svc->value);
 				}
 			}
-			rc_stringlist_free(services);
-			services = NULL;
 		}
+		rc_stringlist_free(services);
+		services = NULL;
 
-		if (tmplist) {
-			TAILQ_FOREACH(svc, tmplist, entries) {
-				if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
+		TAILQ_FOREACH(svc, tmplist, entries) {
+			if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
+				continue;
+			svc_wait(svc->value);
+			if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
+				continue;
+			if (rc_runlevel_stopping()) {
+				/* If shutting down, we should stop even
+				 * if a dependant failed */
+				if (runlevel &&
+				    (strcmp(runlevel,
+					    RC_LEVEL_SHUTDOWN) == 0 ||
+				     strcmp(runlevel,
+					    RC_LEVEL_SINGLE) == 0))
 					continue;
-				svc_wait(svc->value);
-				if (! (rc_service_state(svc->value) & RC_SERVICE_STOPPED)) {
-					if (rc_runlevel_stopping()) {
-						/* If shutting down, we should stop even
-						 * if a dependant failed */
-						if (runlevel &&
-						    (strcmp(runlevel, RC_LEVEL_SHUTDOWN) == 0 ||
-						     strcmp(runlevel, RC_LEVEL_REBOOT) == 0 ||
-						     strcmp(runlevel, RC_LEVEL_SINGLE) == 0))
-							continue;
-						rc_service_mark(service, RC_SERVICE_FAILED);
-					}
-					eerrorx("ERROR: cannot stop %s as %s is still up",
-							applet, svc->value);
-				}
+				rc_service_mark(service, RC_SERVICE_FAILED);
 			}
-			rc_stringlist_free(tmplist);
-			tmplist = NULL;
+			eerrorx("ERROR: cannot stop %s as %s "
+				"is still up", applet, svc->value);
 		}
+		rc_stringlist_free(tmplist);
+		tmplist = NULL;
+	
 
-		/* We now wait for other services that may use us and are stopping
-		   This is important when a runlevel stops */
+		/* We now wait for other services that may use us and are
+		 * stopping. This is important when a runlevel stops */
 		services = rc_deptree_depends(deptree, types_mua, applet_list,
 					      runlevel, depoptions);
-		if (services) {
-			TAILQ_FOREACH(svc, services, entries) {
-				if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
-					continue;
-				svc_wait(svc->value);
-			}
-			rc_stringlist_free(services);
-			services = NULL;
+		TAILQ_FOREACH(svc, services, entries) {
+			if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
+				continue;
+			svc_wait(svc->value);
 		}
+		rc_stringlist_free(services);
+		services = NULL;
 	}
 
 	/* If we're stopping localmount, set LC_ALL=C so that
@@ -1027,10 +1059,10 @@ static void svc_stop(bool deps)
 	if (ibsave)
 		unsetenv("IN_BACKGROUND");
 
-	if (! in_control())
+	if (!in_control())
 		ewarnx("WARNING: %s not under our control, aborting", applet);
 
-	if (! stopped)
+	if (!stopped)
 		eerrorx("ERROR: %s failed to stop", applet);
 
 	if (in_background)
@@ -1046,17 +1078,21 @@ static void svc_stop(bool deps)
 	rc_plugin_run(RC_HOOK_SERVICE_STOP_OUT, applet);
 }
 
-static void svc_restart(bool deps)
+static void
+svc_restart(bool deps)
 {
 	/* This is hairly and a better way needs to be found I think!
-	   The issue is this - openvpn need net and dns. net can restart
-	   dns via resolvconf, so you could have openvpn trying to restart dnsmasq
-	   which in turn is waiting on net which in turn is waiting on dnsmasq.
-	   The work around is for resolvconf to restart it's services with --nodeps
-	   which means just that. The downside is that there is a small window when
-	   our status is invalid.
-	   One workaround would be to introduce a new status, or status locking. */
-	if (! deps) {
+	 * The issue is this - openvpn need net and dns. net can restart
+	 * dns via resolvconf, so you could have openvpn trying to restart
+	 * dnsmasq which in turn is waiting on net which in turn is waiting
+	 * on dnsmasq.
+	 * The work around is for resolvconf to restart it's services with
+	 * --nodeps which means just that.
+	 * The downside is that there is a small window when our status is
+	 * invalid.
+	 * One workaround would be to introduce a new status,
+	 * or status locking. */
+	if (!deps) {
 		RC_SERVICE state = rc_service_state(service);
 		if (state & RC_SERVICE_STARTED || state & RC_SERVICE_INACTIVE)
 			svc_exec("stop", "start");
@@ -1065,7 +1101,7 @@ static void svc_restart(bool deps)
 		return;
 	}
 
-	if (! (rc_service_state(service) & RC_SERVICE_STOPPED)) {
+	if (!(rc_service_state(service) & RC_SERVICE_STOPPED)) {
 		get_started_services();
 		svc_stop(deps);
 	}
@@ -1076,6 +1112,38 @@ static void svc_restart(bool deps)
 	restart_services = NULL;
 }
 
+static bool
+service_plugable(void)
+{
+	char *list, *p, *token;
+	bool allow = true, truefalse;
+	char *match = rc_conf_value("rc_hotplug");
+
+	if (!match)
+		match = rc_conf_value("rc_plug_services");
+	if (!match)
+		return false;
+
+	list = xstrdup(match);
+	p = list;
+	while ((token = strsep(&p, " "))) {
+		if (token[0] == '!') {
+			truefalse = false;
+			token++;
+		} else
+			truefalse = true;
+
+		if (fnmatch(token, applet, 0) == 0) {
+			allow = truefalse;
+			break;
+		}
+	}
+#ifdef DEBUG_MEMORY
+	free(list);
+#endif
+	return allow;
+}
+
 #include "_usage.h"
 #define getoptstring "dDsv" getoptstring_COMMON
 #define extraopts "stop | start | restart | describe | zap"
@@ -1085,7 +1153,7 @@ static const struct option longopts[] = {
 	{ "nodeps",     0, NULL, 'D'},
 	longopts_COMMON
 };
-static const char * const longopts_help[] = {
+static const char *const longopts_help[] = {
 	"set xtrace when running the script",
 	"only run commands when started",
 	"ignore dependencies",
@@ -1093,7 +1161,8 @@ static const char * const longopts_help[] = {
 };
 #include "_usage.c"
 
-int runscript(int argc, char **argv)
+int
+runscript(int argc, char **argv)
 {
 	bool deps = true;
 	bool doneone = false;
@@ -1111,7 +1180,7 @@ int runscript(int argc, char **argv)
 	struct stat stbuf;
 
 	/* Show help if insufficient args */
-	if (argc < 2 || ! exists(argv[1])) {
+	if (argc < 2 || !exists(argv[1])) {
 		fprintf(stderr, "runscript should not be run directly\n");
 		exit(EXIT_FAILURE);
 	}
@@ -1164,20 +1233,6 @@ int runscript(int argc, char **argv)
 	/* Change dir to / to ensure all init scripts don't use stuff in pwd */
 	chdir("/");
 
-#ifdef __linux__
-	/* coldplug events can trigger init scripts, but we don't want to run
-	 * them until after rc sysinit has completed so we punt them to the
-	 * boot runlevel */
-	if (exists("/dev/.rcsysinit")) {
-		eerror("%s: cannot run until sysvinit completes", applet);
-		if (mkdir("/dev/.rcboot", 0755) != 0 && errno != EEXIST)
-			eerrorx("%s: mkdir `/dev/.rcboot': %s", applet, strerror(errno));
-		snprintf(exclusive, sizeof(exclusive), "/dev/.rcboot/%s", applet);
-		symlink(service, exclusive);
-		exit (EXIT_FAILURE);
-	}
-#endif
-
 	if ((runlevel = xstrdup (getenv ("RC_RUNLEVEL"))) == NULL) {
 		env_filter();
 		env_config();
@@ -1197,15 +1252,16 @@ int runscript(int argc, char **argv)
 	if (rc_conf_yesno("rc_parallel")) {
 		/* Get the longest service name */
 		services = rc_services_in_runlevel(NULL);
-		if (services) {
-			TAILQ_FOREACH(svc, services, entries) {
-				ll = strlen(svc->value);
-				if (ll > l)
-					l = ll;
-			}
-			rc_stringlist_free(services);
-			services = NULL;
-		} else l = strlen(applet);
+		TAILQ_FOREACH(svc, services, entries) {
+			ll = strlen(svc->value);
+			if (ll > l)
+				l = ll;
+		}
+		rc_stringlist_free(services);
+		services = NULL;
+		ll = strlen(applet);
+		if (ll > l)
+			l = ll;
 
 		/* Make our prefix string */
 		prefix = xmalloc(sizeof(char) * l + 1);
@@ -1226,13 +1282,14 @@ int runscript(int argc, char **argv)
 	argv++;
 
 	/* Right then, parse any options there may be */
-	while ((opt = getopt_long(argc, argv, getoptstring, longopts, (int *) 0)) != -1)
+	while ((opt = getopt_long(argc, argv, getoptstring,
+				  longopts, (int *)0)) != -1)
 		switch (opt) {
 		case 'd':
 			setenv("RC_DEBUG", "yes", 1);
 			break;
 		case 's':
-			if (! (rc_service_state(service) & RC_SERVICE_STARTED))
+			if (!(rc_service_state(service) & RC_SERVICE_STARTED))
 				exit(EXIT_FAILURE);
 			break;
 		case 'D':
@@ -1250,7 +1307,7 @@ int runscript(int argc, char **argv)
 	}
 
 	if (rc_yesno(getenv("IN_HOTPLUG"))) {
-		if (! rc_conf_yesno("rc_hotplug") || ! service_plugable(applet))
+		if (!service_plugable())
 			eerrorx("%s: not allowed to be hotplugged", applet);
 	}
 
@@ -1285,8 +1342,8 @@ int runscript(int argc, char **argv)
 
 		doneone = true;
 
-		if (strcmp (optarg, "describe") == 0 ||
-		    strcmp (optarg, "help") == 0)
+		if (strcmp(optarg, "describe") == 0 ||
+		    strcmp(optarg, "help") == 0)
 		{
 			save = prefix;
 			eprefix(NULL);
@@ -1303,25 +1360,25 @@ int runscript(int argc, char **argv)
 			   strcmp(optarg, "iprovide") == 0)
 		{
 			errno = 0;
-			if (rc_conf_yesno("rc_depend_strict") || errno == ENOENT)
+			if (rc_conf_yesno("rc_depend_strict") ||
+			    errno == ENOENT)
 				depoptions |= RC_DEP_STRICT;
 
-			if (! deptree && ((deptree = _rc_deptree_load(NULL)) == NULL))
+			if (!deptree &&
+			    ((deptree = _rc_deptree_load(NULL)) == NULL))
 				eerrorx("failed to load deptree");
 
 			tmplist = rc_stringlist_new();
 			rc_stringlist_add(tmplist, optarg);
-			services = rc_deptree_depends(deptree, tmplist, applet_list,
+			services = rc_deptree_depends(deptree, tmplist,
+						      applet_list,
 						      runlevel, depoptions);
 			rc_stringlist_free(tmplist);
-			tmplist = NULL;
-			if (services) {
-				TAILQ_FOREACH(svc, services, entries)
-					printf("%s ", svc->value);
-				printf ("\n");
-				rc_stringlist_free(services);
-				services = NULL;
-			}
+			TAILQ_FOREACH(svc, services, entries)
+				printf("%s ", svc->value);
+			printf ("\n");
+			rc_stringlist_free(services);
+			services = NULL;
 		} else if (strcmp (optarg, "status") == 0) {
 			RC_SERVICE r = svc_status();
 			retval = (int) r;
@@ -1331,7 +1388,8 @@ int runscript(int argc, char **argv)
 			if (strcmp(optarg, "conditionalrestart") == 0 ||
 			    strcmp(optarg, "condrestart") == 0)
 			{
-				if (rc_service_state(service) & RC_SERVICE_STARTED)
+				if (rc_service_state(service) &
+				    RC_SERVICE_STARTED)
 					svc_restart(deps);
 			} else if (strcmp(optarg, "restart") == 0) {
 				svc_restart (deps);
@@ -1340,37 +1398,44 @@ int runscript(int argc, char **argv)
 			} else if (strcmp(optarg, "stop") == 0) {
 				if (deps && in_background)
 					get_started_services();
-
 				svc_stop(deps);
-
 				if (deps) {
 					if (! in_background &&
 					    ! rc_runlevel_stopping() &&
-					    rc_service_state(service) & RC_SERVICE_STOPPED)
-						uncoldplug();
+					    rc_service_state(service) &
+					    RC_SERVICE_STOPPED)
+						unhotplug();
 
 					if (in_background &&
-					    rc_service_state(service) & RC_SERVICE_INACTIVE)
+					    rc_service_state(service) &
+					    RC_SERVICE_INACTIVE)
 					{
-						TAILQ_FOREACH(svc, restart_services, entries)
-							if (rc_service_state(svc->value) & RC_SERVICE_STOPPED)
+						TAILQ_FOREACH(svc,
+							      restart_services, 
+							      entries)
+							if (rc_service_state(svc->value) &
+							    RC_SERVICE_STOPPED)
 								rc_service_schedule_start(service, svc->value);
 					}
 				}
 			} else if (strcmp(optarg, "zap") == 0) {
-				einfo("Manually resetting %s to stopped state", applet);
-				if (!rc_service_mark(applet, RC_SERVICE_STOPPED))
-					eerrorx("rc_service_mark: %s", strerror(errno));
-				uncoldplug();
+				einfo("Manually resetting %s to stopped state",
+				      applet);
+				if (!rc_service_mark(applet,
+						     RC_SERVICE_STOPPED))
+					eerrorx("rc_service_mark: %s",
+						strerror(errno));
+				unhotplug();
 			} else
 				svc_exec(optarg, NULL);
 
-			/* We should ensure this list is empty after an action is done */
+			/* We should ensure this list is empty after
+			 * an action is done */
 			rc_stringlist_free(restart_services);
 			restart_services = NULL;
 		}
 
-		if (! doneone)
+		if (!doneone)
 			usage(EXIT_FAILURE);
 	}
 
