@@ -96,10 +96,10 @@ typedef struct scheduleitem
 	TAILQ_ENTRY(scheduleitem) entries;
 } SCHEDULEITEM;
 TAILQ_HEAD(, scheduleitem) schedule;
-static char **nav = NULL;
+static char **nav;
 
 extern const char *applet;
-static char *changeuser = NULL;
+static char *changeuser, *ch_root, *ch_dir;
 
 extern char **environ;
 
@@ -116,13 +116,14 @@ static void free_schedulelist(void)
 	TAILQ_INIT(&schedule);
 }
 
+#ifdef DEBUG_MEMORY
 static void cleanup(void)
 {
 	free(changeuser);
 	free(nav);
-
 	free_schedulelist();
 }
+#endif
 
 static int parse_signal(const char *sig)
 {
@@ -370,7 +371,7 @@ static int run_stop_schedule(const char *exec, const char *const *argv,
 
 	if (verbose) {
 		if (exec)
-			einfo ("Will stop %s\n", exec);
+			einfo ("Will stop %s", exec);
 		if (pidfile)
 			einfo("Will stop PID in pidfile `%s'", pidfile);
 		if (uid)
@@ -497,6 +498,47 @@ static void handle_signal(int sig)
 	errno = serrno;
 }
 
+static char *
+expand_home(const char *home, const char *path)
+{
+	char *opath, *ppath, *p, *nh;
+	size_t len;
+	struct passwd *pw;
+
+	if (!path || *path != '~')
+		return xstrdup(path);
+
+	opath = ppath = xstrdup(path);
+	if (ppath[1] != '/' && ppath[1] != '\0') {
+		p = strchr(ppath + 1, '/');
+		if (p)
+			*p = '\0';
+		pw = getpwnam(ppath + 1);
+		if (pw) {
+			home = pw->pw_dir;
+			ppath = p;
+			if (ppath)
+				*ppath = '/';
+		} else
+			home = NULL;
+	} else
+		ppath++;
+
+	if (!home) {
+		free(opath);
+		return xstrdup(path);
+	}
+	if (!ppath) {
+		free(opath);
+		return xstrdup(home);
+	}
+
+	len = strlen(ppath) + strlen(home) + 1;
+	nh = xmalloc(len);
+	snprintf(nh, len, "%s%s", home, ppath);
+	free(opath);
+	return nh;
+}
 
 #include "_usage.h"
 #define getoptstring "KN:R:Sbc:d:e:g:k:mn:op:s:tu:r:x:1:2:" getoptstring_COMMON
@@ -562,10 +604,12 @@ int start_stop_daemon(int argc, char **argv)
 #ifdef HAVE_PAM
 	pam_handle_t *pamh = NULL;
 	int pamr;
+	const char *const *pamenv = NULL;
 #endif
 
 	int opt;
-	bool start = true;
+	bool start = false;
+	bool stop = false;
 	bool oknodo = false;
 	bool test = false;
 	bool quiet;
@@ -574,14 +618,13 @@ int start_stop_daemon(int argc, char **argv)
 	char *startas = NULL;
 	char *name = NULL;
 	char *pidfile = NULL;
-	int sig = SIGTERM;
+	int sig = 0;
 	int nicelevel = 0;
 	bool background = false;
 	bool makepidfile = false;
 	uid_t uid = 0;
 	gid_t gid = 0;
-	char *ch_root = NULL;
-	char *ch_dir = NULL;
+	char *home = NULL;
 	int tid = 0;
 	char *redirect_stderr = NULL;
 	char *redirect_stdout = NULL;
@@ -609,7 +652,9 @@ int start_stop_daemon(int argc, char **argv)
 	unsigned int start_wait = 0;
 
 	TAILQ_INIT(&schedule);
+#ifdef DEBUG_MEMORY
 	atexit(cleanup);
+#endif
 
 	signal_setup(SIGINT, handle_signal);
 	signal_setup(SIGQUIT, handle_signal);
@@ -620,16 +665,24 @@ int start_stop_daemon(int argc, char **argv)
 			eerror("%s: invalid nice level `%s' (SSD_NICELEVEL)",
 				applet, tmp);
 
+	/* Get our initial dir */
+	home = getenv("HOME");
+	if (!home) {
+		pw = getpwuid(getuid());
+		if (pw)
+			home = pw->pw_dir;
+	}
+
 	while ((opt = getopt_long(argc, argv, getoptstring, longopts,
 				  (int *) 0)) != -1)
 		switch (opt) {
 		case 'K':  /* --stop */
-			start = false;
+			stop = true;
 			break;
-
 		case 'N':  /* --nice */
 			if (sscanf(optarg, "%d", &nicelevel) != 1)
-				eerrorx("%s: invalid nice level `%s'", applet, optarg);
+				eerrorx("%s: invalid nice level `%s'",
+					applet, optarg);
 			break;
 
 		case 'R':  /* --retry <schedule>|<timeout> */
@@ -655,10 +708,12 @@ int start_stop_daemon(int argc, char **argv)
 				else
 					pw = getpwuid((uid_t) tid);
 
-				if (! pw)
-					eerrorx("%s: user `%s' not found", applet, tmp);
+				if (!pw)
+					eerrorx("%s: user `%s' not found",
+						applet, tmp);
 				uid = pw->pw_uid;
-				if (! gid)
+				home = pw->pw_dir;
+				if (!gid)
 					gid = pw->pw_gid;
 
 				if (p) {
@@ -668,8 +723,9 @@ int start_stop_daemon(int argc, char **argv)
 					else
 						gr = getgrgid((gid_t) tid);
 
-					if (! gr)
-						eerrorx("%s: group `%s' not found",
+					if (!gr)
+						eerrorx("%s: group `%s'"
+							" not found",
 							applet, tmp);
 					gid = gr->gr_gid;
 				}
@@ -682,9 +738,10 @@ int start_stop_daemon(int argc, char **argv)
 
 		case 'e': /* --env */
 			if (putenv(optarg) == 0) {
-				if (strncmp("HOME=", optarg, 5) == 0)
+				if (strncmp("HOME=", optarg, 5) == 0) {
 					sethome = true;
-				else if (strncmp("USER=", optarg, 5) == 0)
+					home = strchr(optarg, '=') + 1;
+				} else if (strncmp("USER=", optarg, 5) == 0)
 					setuser = true;
 			}
 			break;
@@ -696,8 +753,9 @@ int start_stop_daemon(int argc, char **argv)
 				else
 					gr = getgrgid((gid_t) tid);
 
-				if (! gr)
-					eerrorx("%s: group `%s' not found", applet, optarg);
+				if (!gr)
+					eerrorx("%s: group `%s' not found",
+						applet, optarg);
 				gid = gr->gr_gid;
 			}
 			break;
@@ -764,6 +822,7 @@ int start_stop_daemon(int argc, char **argv)
 	/* Allow start-stop-daemon --signal HUP --exec /usr/sbin/dnsmasq
 	 * instead of forcing --stop --oknodo as well */
 	if (!start &&
+	    !stop &&
 	    sig != SIGINT &&
 	    sig != SIGTERM &&
 	    sig != SIGQUIT &&
@@ -777,31 +836,44 @@ int start_stop_daemon(int argc, char **argv)
 
 	if (!exec) {
 		exec = *argv;
-		if (name)
+		if (!exec)
+			exec = name;
+		if (name && start)
 			*argv = name;
-	} else if (name)
+	} else if (name && (start || **argv))
 		*--argv = name;
 	else
 		*--argv = exec;
 
+	if (stop || sig) {
+		if ( !*argv && !pidfile && !name && !uid)
+			eerrorx("%s: --stop needs --exec, --pidfile,"
+				" --name or --user", applet);
+		if (background)
+			eerrorx("%s: --background is only relevant with"
+				" --start", applet);
+		if (makepidfile)
+			eerrorx("%s: --make-pidfile is only relevant with"
+				" --start", applet);
+		if (redirect_stdout || redirect_stderr)
+			eerrorx("%s: --stdout and --stderr are only relevant"
+				" with --start", applet);
+	} else {
+		if (!exec)
+			eerrorx("%s: nothing to start", applet);
+		if (makepidfile && !pidfile)
+			eerrorx("%s: --make-pidfile is only relevant with"
+				" --pidfile", applet);
+		if ((redirect_stdout || redirect_stderr) && !background)
+			eerrorx("%s: --stdout and --stderr are only relevant"
+				" with --background", applet);
+	}
 
-	if (start && !exec)
-		eerrorx("%s: nothing to start", applet);
-
-	if (!start && !*argv && !pidfile && !name && !uid)
-		eerrorx("%s: --stop needs --exec, --pidfile, --name or --user", applet);
-
-	if (makepidfile && !pidfile)
-		eerrorx("%s: --make-pidfile is only relevant with --pidfile", applet);
-
-	if (background && !start)
-		eerrorx("%s: --background is only relevant with --start", applet);
-
-	if ((redirect_stdout || redirect_stderr) && !background)
-		eerrorx("%s: --stdout and --stderr are only relevant with --background",
-			 applet);
-
-	if (!start) {
+	if (stop || sig) {
+		if (!sig)
+			sig = SIGTERM;
+		if (!stop)
+			oknodo = true;
 		if (!TAILQ_FIRST(&schedule)) {
 			if (test || oknodo)
 				parse_schedule("0", sig);
@@ -830,6 +902,14 @@ int start_stop_daemon(int argc, char **argv)
 		exit(EXIT_SUCCESS);
 	}
 
+	/* Expand ~ */
+	if (ch_dir && *ch_dir == '~')
+		ch_dir = expand_home(home, ch_dir);
+	if (ch_root && *ch_root == '~')
+		ch_root = expand_home(home, ch_root);
+	if (*exec == '~')
+		exec = expand_home(home, exec);
+
 	/* Validate that the binary exists if we are starting */
 	if (*exec == '/' || *exec == '.') {
 		/* Full or relative path */
@@ -992,9 +1072,9 @@ int start_stop_daemon(int argc, char **argv)
 			eerrorx("%s: unable to set groupid to %d", applet, gid);
 		if (changeuser && initgroups(changeuser, gid))
 			eerrorx("%s: initgroups (%s, %d)", applet, changeuser, gid);
-		if (uid && setuid(uid))
-			eerrorx ("%s: unable to set userid to %d", applet, uid);
-		else {
+		if (uid) {
+			if (setuid(uid))
+				eerrorx ("%s: unable to set userid to %d", applet, uid);
 			pw = getpwuid(uid);
 			if (pw) {
 				if (!sethome) {
@@ -1023,6 +1103,19 @@ int start_stop_daemon(int argc, char **argv)
 		i = 0;
 		while(environ[i])
 			rc_stringlist_add(env_list, environ[i++]);
+
+#ifdef HAVE_PAM
+		pamenv = (const char *const *)pam_getenvlist(pamh);
+		if (pamenv) {
+			while (*pamenv) {
+				/* Don't add strings unless they set a var */
+				if (strchr(*pamenv, '='))
+					putenv(xstrdup(*pamenv));
+				pamenv++;
+			}
+		}
+#endif
+
 		TAILQ_FOREACH(env, env_list, entries) {
 			if ((strncmp(env->value, "RC_", 3) == 0 &&
 			     strncmp(env->value, "RC_SERVICE=", 10) != 0 &&
@@ -1038,22 +1131,28 @@ int start_stop_daemon(int argc, char **argv)
 		rc_stringlist_free(env_list);
 
 		/* For the path, remove the rcscript bin dir from it */
-		if ((tmp = xstrdup(getenv("PATH")))) {
-			len = strlen(tmp);
-			newpath = np = xmalloc(len);
-			p = tmp;
-			while ((token = strsep(&p, ":"))) {
-				if (strcmp(token, RC_LIBDIR "/bin") == 0 ||
-				    strcmp(token, RC_LIBDIR "/sbin") == 0)
-					continue;
-				len = strlen(token);
-				if (np != newpath)
-					*np++ = ':';
-				memcpy(np, token, len);
-				np += len;
-				*np = '\0';
+		if ((token = getenv("PATH"))) {
+			len = strlen(token);
+			newpath = np = xmalloc(len + 1);
+			while (token && *token) {
+				p = strchr(token, ':');
+				if (p) {
+					*p++ = '\0';
+					while (*p == ':')
+						p++;
+				}
+				if (strcmp(token, RC_LIBDIR "/bin") != 0 &&
+				    strcmp(token, RC_LIBDIR "/sbin") != 0)
+				{
+					len = strlen(token);
+					if (np != newpath)
+						*np++ = ':';
+					memcpy(np, token, len);
+					np += len;
+				}
+				token = p;
 			}
-			free(tmp);
+			*np = '\0';
 			unsetenv("PATH");
 			setenv("PATH", newpath, 1);
 		}
